import os
import shutil


def rm_rf(path):
    path = str(path)
    if os.path.isdir(path) and not os.path.islink(path):
        shutil.rmtree(path)
    elif os.path.exists(path):
        os.remove(path)



import multiprocessing

import tqdm


class parallel_tqdm(tqdm.tqdm):

    def __init__(self, *args, **kwargs):
        super(parallel_tqdm, self).__init__(*args, **kwargs)
        self.correct_count = multiprocessing.Value('i', 0)

    def update(self, n=1):
        with self.correct_count.get_lock():
            self.n = self.correct_count.value
            super(parallel_tqdm, self).update(n)
            self.correct_count.value += n

    def close(self):
        if self.disable:
            return
        self.n = self.correct_count.value
        super(parallel_tqdm, self).close()


if __name__ == '__main__':
    a = range(1000)
    with parallel_tqdm(a) as pbar:
        for _i in a:
            pbar.update()
    pbar.close()



#!/usr/bin/env python

import argparse

from pybsc import load_json
from pybsc import save_json


def main():
    parser = argparse.ArgumentParser(description='Reformat json')
    parser.add_argument('input_json', type=str, help='Input json file')
    parser.add_argument('output_json', type=str, help='Output json file')
    args = parser.parse_args()

    data = load_json(args.input_json)
    save_json(data, args.output_json)


if __name__ == '__main__':
    main()



import contextlib
import wave


def get_wave_duration(wave_filepath):
    with contextlib.closing(wave.open(str(wave_filepath), 'r')) as f:
        frames = f.getnframes()
        rate = f.getframerate()
        duration = frames / float(rate)
    return duration




from collections import defaultdict

import numpy as np
from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont

from pybsc.data import font_path

category_trans = int(0.6 * 255)


def visualize_bboxes(img, bboxes, labels=None, copy=False,
                     font_size=None, box_color=(191, 40, 41),
                     box_alpha=255,
                     text_alpha=None,
                     background_alpha=None,
                     font_ratio=20,
                     text_colors=None,
                     box_colors=None,
                     box_alphas=None):
    """Visualize bounding boxes.

    Note that if alpha is 0, the target color will be transparent.
    """
    height, width, _ = img.shape
    long_side = min(width, height)
    font_size = font_size or max(int(round(long_side / font_ratio)), 1)
    box_width = max(int(round(long_side / 180)), 1)
    font = ImageFont.truetype(font_path(), font_size)

    result_vis = Image.new("RGBA", (width, height), (0, 0, 0, 0))
    draw = ImageDraw.Draw(result_vis)

    for i, (_, (y1, x1, y2, x2)) in enumerate(bboxes):
        if box_alphas is not None:
            box_alpha = box_alphas[i]
        if box_colors is not None:
            draw.rectangle((x1, y1, x2, y2),
                           outline=box_colors[i] + (box_alpha,),
                           width=box_width)
        else:
            draw.rectangle((x1, y1, x2, y2), outline=box_color + (box_alpha,),
                           width=box_width)

    bbox_captions = defaultdict(list)
    if labels is not None:
        if len(labels) > 0:
            if isinstance(labels[0], str):
                for i_box, label in enumerate(labels):
                    bbox_captions[i_box].append((label, box_color, 'left top'))
            elif len(labels[0]) == 3:
                for i_box, label, color in labels:
                    bbox_captions[i_box].append((label, color, 'left top'))
            elif len(labels[0]) == 4:
                for i_box, label, color, loc in labels:
                    bbox_captions[i_box].append((label, color, loc))

    alpha = background_alpha if background_alpha is not None \
        else category_trans
    text_alpha = text_alpha if text_alpha is not None else category_trans
    text_color = (255, 255, 255)
    for i_box, (score, box) in enumerate(bboxes):
        if box_alphas is not None and box_alphas[i_box] == 0:
            continue

        captions = []
        bg_colors = []
        locs = []

        loc = 'left top'
        for label, color, loc in bbox_captions[i_box]:
            captions.append(label)
            bg_colors.append(color)
            locs.append(loc)

        if score is not None:
            conf = f" {score:.2f}"
            captions.append('score ' + conf)
            bg_colors.append((176, 85, 234))
            locs.append(loc)

        if len(captions) == 0:
            continue
        y1, x1, y2, x2 = box
        overlay = Image.new("RGBA", result_vis.size, (0, 0, 0, 0))
        trans_draw = ImageDraw.Draw(overlay)
        caption_sizes = [trans_draw.textsize(caption, font=font)
                         for caption in captions]
        caption_widths, caption_heights = list(zip(*caption_sizes))
        max_height = max(caption_heights)
        rec_height = int(round(1.8 * max_height))
        space_height = int(round(0.2 * max_height))
        total_height = (rec_height + space_height) \
            * (len(captions) - 1) \
            + rec_height
        width_pad = max(font_size // 2, 1)
        start_y = max(round(y1) - total_height, space_height)

        right_i = 0
        left_i = 0
        for i, (caption, loc) in enumerate(zip(captions, locs)):
            height_pad = round((rec_height - caption_heights[i]) / 2)
            if loc == 'left top':
                r_x1 = round(x1)
                r_y1 = start_y + (rec_height + space_height) * left_i
                r_x2 = r_x1 + caption_widths[i] + width_pad * 2
                r_y2 = r_y1 + rec_height
                left_i += 1
            elif loc == 'right top':
                r_y1 = start_y + (rec_height + space_height) * right_i
                r_x1 = x2 - (caption_widths[i] + width_pad * 2)
                r_y2 = r_y1 + rec_height
                r_x2 = x2
                right_i += 1
            else:
                raise RuntimeError
            rec_pos = (r_x1, r_y1, r_x2, r_y2)
            text_pos = (r_x1 + width_pad, r_y1 + height_pad)

            bg_color_alpha = alpha
            if box_alphas is not None:
                bg_color_alpha = box_alphas[i_box]
            trans_draw.rectangle(rec_pos, fill=bg_colors[i]
                                 + (int(bg_color_alpha),))
            # TODO(iory) Fix text color.
            if np.linalg.norm(
                    np.array(bg_colors[i]) - np.array(text_color)) <= 40:
                tc = (0, 0, 0)
            else:
                tc = text_color
            if box_alphas is not None:
                text_alpha = box_alphas[i_box]
            trans_draw.text(text_pos, caption,
                            fill=tc + (text_alpha,), font=font,
                            align="center")
        result_vis = Image.alpha_composite(result_vis, overlay)

    pil_img = Image.fromarray(img[..., ::-1])
    pil_img = pil_img.convert("RGBA")
    pil_img = Image.alpha_composite(pil_img, result_vis)
    pil_img = pil_img.convert("RGB")
    if copy:
        return np.array(pil_img, dtype=np.uint8)[..., ::-1]
    else:
        img[:] = np.array(pil_img, dtype=np.uint8)[..., ::-1]
        return img



import functools
from numbers import Integral

# RGB order
flat_colors = {
    'red': (231, 76, 60),
    'blue': (52, 152, 219),
    'green': (46, 204, 113),
    'yellow': (241, 196, 15),
    'gray': (149, 165, 166),
    'silver': (189, 195, 199),
    'midnight_blue': (44, 62, 80),
    'purple': (142, 68, 173),
    'orange': (243, 156, 18),
    'turquoise': (26, 188, 156),
    'white': (245, 246, 250),
    'black': (47, 54, 64),
    'violet': (155, 89, 182),
    'brown': (93, 64, 55),
}


@functools.lru_cache(maxsize=None)
def _voc_colormap(i):
    r, g, b = 0, 0, 0
    for j in range(8):
        if i & (1 << 0):
            r |= 1 << (7 - j)
        if i & (1 << 1):
            g |= 1 << (7 - j)
        if i & (1 << 2):
            b |= 1 << (7 - j)
        i >>= 3
    return r, g, b


def voc_colormap(nlabels, order='rgb'):
    colors = []
    for i in range(nlabels):
        r, g, b = _voc_colormap(i)
        if order == 'rgb':
            colors.append([r, g, b])
        elif order == 'bgr':
            colors.append([b, g, r])
    return colors


def cmap(value):
    if isinstance(value, Integral):
        return _voc_colormap(value)
    if isinstance(value, str):
        return flat_colors[value]
    if isinstance(value, tuple):
        return value



import subprocess

import six


def run_command(cmd, *args, **kwargs):
    """Execute a command.

    Parameters
    ----------
    cmd : str or list of str
        The command to execute. For a single command, it can be a string.
        For multiple commands, it should be a list of strings.
    *args
        Additional positional arguments to be passed to the subprocess call.
    **kwargs
        Additional keyword arguments to be passed to the subprocess call.
        capture_output : bool, optional
            If True, capture the output of the command (stdout and stderr)
            and return it. By default, the output is not captured.

    Returns
    -------
    CompletedProcess or int
        If `capture_output` is True, returns a `CompletedProcess` object
        containing information about the completed process, including the
        captured output. If `capture_output` is False, returns the process
        return code as an integer.

    Raises
    ------
    CalledProcessError
        If the command execution fails and raises an error.

    Notes
    -----
    This function uses the `subprocess` module to execute the command.
    On Python 2, it uses `subprocess.check_call`, and on Python 3 or later,
    it uses `subprocess.run` to run the command.

    Example
    -------
    >>> run_command('ls -l')
    # Executes the 'ls -l' command and returns the process return code.

    >>> run_command(['git', 'status'], capture_output=True)
    # Executes the 'git status' command, captures the output
    # (stdout and stderr),
    # and returns a CompletedProcess object with the captured output.
    """
    if kwargs.pop("capture_output", False):
        kwargs["stdout"] = subprocess.PIPE
        kwargs["stderr"] = subprocess.PIPE
    if six.PY2:
        return subprocess.check_call(cmd, *args, **kwargs)
    else:
        return subprocess.run(cmd, *args, **kwargs)



import hashlib


def checksum_md5(filename, blocksize=8192):
    """Calculate md5sum.

    Parameters
    ----------
    filename : str or pathlib.Path
        input filename.
    blocksize : int
        MD5 has 128-byte digest blocks (default: 8192 is 128x64).

    Returns
    -------
    md5 : str
        calculated md5sum.
    """
    filename = str(filename)
    hash_factory = hashlib.md5()
    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(blocksize), b''):
            hash_factory.update(chunk)
    return hash_factory.hexdigest()




import datetime
import math
import multiprocessing
from multiprocessing import Pool
import os
import os.path as osp
from pathlib import Path
import shutil
import subprocess
import tempfile
from tempfile import NamedTemporaryFile

import cv2
import skvideo.io
from tqdm import tqdm

from pybsc.subprocess_utils import run_command
from pybsc.timestamp_utils import get_timestamp_indices_wrt_base


def hhmmss_to_seconds(ts):
    try:
        return float(ts)
    except ValueError:
        return sum(
            float(x) * 60 ** i
            for i, x in enumerate(reversed(ts.split(':'))))


def get_video_aspect_ratio(video_path):
    w, h = get_video_size(video_path)
    gcd = math.gcd(w, h)
    return w // gcd, h // gcd


def get_video_duration(video_path):
    video_path = str(video_path)
    if not osp.exists(video_path):
        raise OSError(f"{video_path} not exists")
    metadata = skvideo.io.ffprobe(video_path)
    return float(metadata['video']['@duration'])


def get_video_avg_frame_rate(video_path):
    video_path = str(video_path)
    if not osp.exists(video_path):
        raise OSError(f"{video_path} not exists")
    metadata = skvideo.io.ffprobe(video_path)
    a, b = metadata['video']['@avg_frame_rate'].split('/')
    a = int(a)
    b = int(b)
    return a / b


def get_video_size(video_path):
    video_path = str(video_path)
    if not osp.exists(video_path):
        raise OSError(f"{video_path} not exists")
    metadata = skvideo.io.ffprobe(video_path)
    height = int(metadata['video']['@height'])
    width = int(metadata['video']['@width'])
    return width, height


def get_video_n_frame(video_path):
    video_path = str(video_path)
    if not osp.exists(video_path):
        raise OSError(f"{video_path} not exists")
    metadata = skvideo.io.ffprobe(video_path)
    if '@nb_frames' not in metadata['video']:
        fps = get_video_avg_frame_rate(video_path)
        return int(fps * get_video_duration(video_path))
    return int(metadata['video']['@nb_frames'])


def get_video_creation_time(video_path):
    metadata = skvideo.io.ffprobe(video_path)
    tag_dict = {}
    for tag in metadata['video']['tag']:
        tag_dict[tag['@key']] = tag['@value']
    if 'creation_time' not in tag_dict:
        return None
    creation_time = tag_dict['creation_time']
    created_at = datetime.datetime.strptime(
        creation_time, '%Y-%m-%dT%H:%M:%S.%fZ')
    return created_at


def split_video(video_file_path, output_path=None,
                start_time=None, end_time=None,
                hflip=False, vflip=False,
                out_to_null=True):
    if start_time is None and end_time is None:
        raise ValueError
    start_time = start_time or 0.0
    end_time = end_time or get_video_duration(video_file_path)
    command = ['ffmpeg',
               '-y',
               '-ss', str(start_time),
               '-t', str(end_time - start_time),
               '-i', str(video_file_path)]
    if hflip is True:
        command.extend(['-vf', 'hflip'])
    if vflip is True:
        command.extend(['-vf', 'vflip'])

    stdout = None
    stderr = None
    if out_to_null:
        stdout = subprocess.DEVNULL
        stderr = subprocess.DEVNULL
    if output_path is None:
        suffix = Path(video_file_path).suffix
        with NamedTemporaryFile(delete=True, suffix=suffix) as tmp:
            command.append(tmp.name)
            if not out_to_null:
                print(f'Running \"{" ".join(command)}\"')
            proc = subprocess.Popen(command,
                                    stdout=stdout,
                                    stderr=stderr)
            proc.wait()
            shutil.copy(tmp.name, video_file_path)
    else:
        command.append(str(output_path))
        if not out_to_null:
            print(f'Running \"{" ".join(command)}\"')
        proc = subprocess.Popen(command,
                                stdout=stdout,
                                stderr=stderr)
        proc.wait()


def split_video_from_stamps(
        video_file_path,
        start_stamps, end_stamps,
        output_paths=None,
        output_dirpath=None,
        filename_tmpl='video_{start_frame:05}_{end_frame:05}.mp4',
        show_progress=True,
        parallel_process=False,
        n_jobs=8,
        hflip=False, vflip=False):
    if len(start_stamps) != len(end_stamps):
        raise RuntimeError(
            + 'Length not much len(start_stamps) != len(end_stamps)'
            + f'({len(start_stamps)})!=({len(end_stamps)})')
    fps = get_video_avg_frame_rate(video_file_path)
    if output_paths is None:
        output_paths = []
        if output_dirpath is None:
            tmp_output_dir = tempfile.TemporaryDirectory()
            tmp_output_dir.cleanup()
            os.makedirs(tmp_output_dir.name, exist_ok=True)
            output_dirpath = tmp_output_dir.name
        output_dirpath = Path(output_dirpath)
        for _i, (s, e) in enumerate(zip(start_stamps, end_stamps)):
            start_frame = int(math.ceil(s * fps))
            end_frame = int(math.ceil(e * fps))
            output_paths.append(output_dirpath
                                / filename_tmpl.format(
                                    start_frame=start_frame,
                                    end_frame=end_frame))
    if parallel_process is False:
        for _i, (s, e, video_out) in tqdm(enumerate(zip(
                start_stamps, end_stamps, output_paths)),
                                         total=len(output_paths),
                                         disable=show_progress is False):
            split_video(video_file_path, video_out, start_time=s, end_time=e,
                        hflip=hflip, vflip=vflip)
    else:
        cpu_count = multiprocessing.cpu_count()
        if n_jobs == -1 or n_jobs == 'full':
            n_jobs = min(cpu_count, len(output_paths))
        elif n_jobs == 'half':
            n_jobs = min(cpu_count // 2, len(output_paths))
        p = Pool(n_jobs)
        for _i, (s, e, video_out) in enumerate(zip(
                start_stamps, end_stamps, output_paths)):
            p.apply_async(split_video,
                          args=(video_file_path, video_out, s, e,
                                hflip, vflip))
        p.close()
        p.join()
    return output_paths


def extract_target_frame_from_timestamp(video_filepath, timestamp):
    """Extract target frame from timestamp.

    Parameters
    ----------
    video_filepath : str
        video filepath
    timestamp : float
        timestamp in seconds.

    Returns
    -------
    image : numppy.ndarray
    """
    duration = get_video_duration(video_filepath)
    timestamp = max(0.0, min(duration, timestamp))
    vidcap = cv2.VideoCapture(str(video_filepath))
    vidcap.set(cv2.CAP_PROP_POS_MSEC, int(timestamp * 1000))
    success, image = vidcap.read()
    return image


def calc_frame_to_second_coefficient(duration, n_frame, sampling_sec=None):
    """Calculate Frame to Second coefficient.

    Parameters
    ----------
    duration : float
        video duration
    n_frame : int
        number of frame
    sampling_sec : float
        time of sampling seconds

    Returns
    -------
    frame_to_second : float
        calculated frame to second coefficient.

    Examples
    --------
    >>> calc_frame_to_second_coefficient(30, 60, sampling_sec=0.1)
    0.5
    >>> calc_frame_to_second_coefficient(30, 60, sampling_sec=0.1) * 3
    1.5
    """
    duration = float(duration)
    n_frame = float(n_frame)
    fps = n_frame / duration
    if sampling_sec is not None:
        return duration * math.ceil(fps * sampling_sec) / n_frame
    else:
        return duration / n_frame


def calc_second_to_frame_coefficient(duration, n_frame, sampling_sec=None):
    """Calculate Second to Frame coefficient.

    This is inverse of calc_frame_to_second_coefficient.

    Parameters
    ----------
    duration : float
        video duration
    n_frame : int
        number of frame
    sampling_sec : float
        time of sampling seconds

    Returns
    -------
    second_to_frame : float
        calculated second to frame coefficient.

    Examples
    --------
    >>> calc_second_to_frame_coefficient(30, 60, sampling_sec=0.1)
    2.0
    >>> int(1.99 * calc_second_to_frame_coefficient(4, 4))
    1
    >>> int(1.99 * calc_second_to_frame_coefficient(4, 4, sampling_sec=2.0))
    0
    """
    return 1.0 / calc_frame_to_second_coefficient(
        duration, n_frame, sampling_sec)


def load_frame(video_path, start=0.0, duration=-1,
               target_size=None, sampling_frequency=None):
    """Load frame

    Parameters
    ----------
    video_path : str or pathlib.Path
        input video path.
    start : float
        start time
    duration : int or float
        duration. If this value is `-1`, load all frames.

    Returns
    -------
    frames : list[numpy.ndarray]
        all frames.
    stamps : list[float]
        time stamps.
    """
    video_path = str(video_path)
    vid = cv2.VideoCapture(video_path)
    fps = vid.get(cv2.CAP_PROP_FPS)
    vid.set(cv2.CAP_PROP_POS_MSEC, start)
    vid_avail = True
    if sampling_frequency is not None:
        frame_interval = int(math.ceil(fps * sampling_frequency))
    else:
        frame_interval = 1
    cur_frame = 0
    while True:
        stamp = float(cur_frame) / fps
        vid_avail, frame = vid.read()
        if not vid_avail:
            break
        if duration != -1 and stamp > start + duration:
            break
        if target_size is not None:
            frame = cv2.resize(frame, target_size)
        yield frame, stamp
        cur_frame += frame_interval
        vid.set(cv2.CAP_PROP_POS_FRAMES, cur_frame)
    vid.release()


def count_frames(video_path, start=0.0, duration=-1,
                 sampling_frequency=None):
    video_duration = get_video_duration(video_path)
    video_duration -= start
    if duration > 0:
        video_duration = max(video_duration - duration, 0)
    fps = get_video_avg_frame_rate(video_path)
    if sampling_frequency is not None:
        return int(math.ceil(
            video_duration * fps
            / int(math.ceil(fps * sampling_frequency))))
    else:
        return int(math.ceil(video_duration * fps))


def video2fileframes(
        video_path,
        output_path=None,
        filename_tmpl='img_{:05}.jpg'):
    if output_path is None:
        tmp_output_dir = tempfile.TemporaryDirectory()
        tmp_output_dir.cleanup()
        os.makedirs(tmp_output_dir.name, exist_ok=True)
        output_path = tmp_output_dir.name
    output_path = Path(output_path)
    outfiles = []
    timestamps = []
    for i, (frame, stamp) in enumerate(
            load_frame(video_path), 0):
        outfile = output_path / filename_tmpl.format(i)
        cv2.imwrite(str(outfile), frame)
        outfiles.append(outfile)
        timestamps.append(stamp)
    return outfiles, timestamps


def split_video_to_fileframe_from_sections(
        outfiles, timestamps, start_stamps, end_stamps,
        output_path=None,
        filename_tmpl='img_{:05}.jpg',
        dirname_tmpl='frames_{:05}-{:05}'):
    if output_path is None:
        tmp_output_dir = tempfile.TemporaryDirectory()
        tmp_output_dir.cleanup()
        os.makedirs(tmp_output_dir.name, exist_ok=True)
        output_path = tmp_output_dir.name
    output_path = Path(output_path)
    output_frame_parent_paths = []
    for start_stamp, end_stamp in zip(start_stamps, end_stamps):
        start = get_timestamp_indices_wrt_base(timestamps, start_stamp)
        end = get_timestamp_indices_wrt_base(timestamps, end_stamp)
        frame_outpath = output_path / filename_tmpl.format(start, end)
        os.makedirs(str(frame_outpath), exist_ok=True)
        for i, filename in enumerate(outfiles[start:end]):
            filename = Path(filename)
            shutil.copy(filename, frame_outpath / filename_tmpl.format(i))
        output_frame_parent_paths.append(frame_outpath)
    return output_frame_parent_paths


def get_video_rotation(video_file_path):
    """Function to get the rotation of the input video file.

    Returns a rotation None, 90, 180 or 270
    """
    cmd = "ffprobe -loglevel error -select_streams v:0 -show_entries stream_tags=rotate -of default=nw=1:nk=1 {}".format(video_file_path)  # NOQA
    ffprobe_output = run_command(cmd, shell=True, capture_output=True)
    # Output of cmdis None if it should be 0
    if len(ffprobe_output.stdout) > 0:
        rotation = int(ffprobe_output.stdout)
    else:
        rotation = 0
    return rotation



from itertools import tee

try:
    from itertools import zip_longest
except Exception:
    from itertools import izip_longest as zip_longest


def pairwise(iterable, longest=False):
    a, b = tee(iterable, 2)
    next(b, None)
    if longest:
        return zip_longest(a, b)
    else:
        return zip(a, b)


def triplewise(iterable, longest=False):
    a, b, c = tee(iterable, 3)
    next(b, None)
    next(c, None)
    next(c, None)
    if longest:
        return zip_longest(a, b, c)
    else:
        return zip(a, b, c)



import os
import sys


class suppress_stdout:

    """A context manager that block stdout for its scope

    with HideOutput():
        os.system('ls -l')

    From https://stackoverflow.com/a/17954769/4176597

    """

    def __init__(self, to=os.devnull):
        sys.stdout.flush()
        self._org_stdout = sys.stdout
        self._old_stdout_fno = os.dup(sys.stdout.fileno())
        self._to = os.open(to, os.O_WRONLY)

    def __enter__(self):
        self._newstdout = os.dup(1)
        os.dup2(self._to, 1)
        os.close(self._to)
        sys.stdout = os.fdopen(self._newstdout, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self._org_stdout
        sys.stdout.flush()
        os.dup2(self._old_stdout_fno, 1)



def _is_sorted_numpy(x):
    import numpy as np
    return np.all(np.diff(x) >= 0)


def _is_sorted(x):
    return sorted(x) == x


def is_sorted(x):
    if isinstance(x, list) or isinstance(x, tuple):
        return _is_sorted(x)
    elif x.__class__.__name__ == 'ndarray':
        return _is_sorted_numpy(x)
    else:
        raise NotImplementedError



import numpy as np


def squared_tile(n):
    if n < 1:
        raise ValueError('input n should be greater than 0.')
    tmp = int(np.ceil(np.sqrt(n)))
    if tmp * tmp == n:
        return (tmp, tmp)
    if tmp * (tmp - 1) >= n:
        return tmp, tmp - 1
    i = 0
    while tmp * (tmp - i) > n:
        i += 1
    return (tmp, tmp - i + 1)



from numbers import Number

import numpy as np


def _find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx


def argclosest(array, values):
    if isinstance(values, Number):
        return _find_nearest(array, values)

    # https://stackoverflow.com/questions/2566412/
    # find-nearest-value-in-numpy-array
    array = np.array(array)

    # get insert positions
    idxs = np.searchsorted(array, values, side="left")

    # find indexes where previous index is closer
    prev_idx_is_less = (
        (idxs == len(array))
        | (np.fabs(values - array[np.maximum(idxs - 1, 0)])
           < np.fabs(values - array[np.minimum(idxs, len(array) - 1)])))
    idxs[prev_idx_is_less] -= 1
    return idxs



# flake8: noqa

import pkg_resources


__version__ = pkg_resources.get_distribution("pybsc").version

from pybsc.compress import make_tarfile
from pybsc.make_fancy_output_dir import make_fancy_output_dir
from pybsc.makedirs import makedirs
from pybsc.measure import measure
from pybsc.run_many import run_many
from pybsc.time_str import current_time_str
from pybsc.which import which
from pybsc.git import touch_gitignore

from pybsc.split import nsplit

from pybsc.iter_utils import pairwise
from pybsc.iter_utils import triplewise

from pybsc.json_utils import load_json
from pybsc.json_utils import save_json

from pybsc.yaml_utils import load_yaml
from pybsc.yaml_utils import save_yaml

from pybsc.dict_utils import invert_dict

from pybsc.download import download_file

from pybsc.sort_utils import is_sorted

from pybsc.pickle_utils import load_pickle

from pybsc.rm_utils import rm_rf

from pybsc.cache_readline_history import cache_readline

from pybsc.closest_indices import argclosest

from pybsc.is_nan import is_nan

from pybsc.get_target_ext import get_target_ext

import pybsc.parallel

import pybsc.heatmap

from pybsc.opener import opener

from pybsc.md5sum_utils import checksum_md5

from pybsc.stdout_utils import suppress_stdout

from pybsc.str_utils import removesuffix
from pybsc.str_utils import remove_non_ascii

from pybsc.bytes_utils import bytes_to_uint8_list

from pybsc.subprocess_utils import run_command

from pybsc.cli import query_yes_no



import atexit
import readline


def cache_readline(histfile, histlength=1000, emacs=True):
    try:
        readline.read_history_file(histfile)
        readline.set_history_length(histlength)
    except FileNotFoundError:
        pass
    if emacs is True:
        readline.parse_and_bind("set editing-mode emacs")
    atexit.register(readline.write_history_file, histfile)



import numpy as np

from pybsc.closest_indices import argclosest


def get_timestamp_indices_wrt_base(base_timestamps, target_timestamps,
                                   style='after'):
    """Return timestamp indices with respect to base timestamps.

    Examples
    --------
    >>> base_timestamps = np.array([0., 1., 2., 3., 4., 5.])
    >>> target_timestamps = np.array([0.21, 0.43, 1.4, 3.1])
    >>> get_timestamp_indices_wrt_base(base_timestamps, target_timestamps)
    array([0, 0, 1, 3])
    """
    if style == 'after':
        return np.maximum(
            np.searchsorted(base_timestamps,
                            target_timestamps, side="left") - 1, 0)
    elif style == 'closest':
        return argclosest(base_timestamps, target_timestamps)
    else:
        raise NotImplementedError


def sorted_timestamps(timestamps):
    """Returns a timestamp that is sorted and whose time starts at 0 seconds.

    Examples
    --------
    >>> timestamps = np.array([5, 3, 2, 1])
    >>> sorted_timestamps(timestamps)
    array([0, 1, 2, 4])
    """
    np_stamps = np.array(timestamps, dtype=np.float64)
    np_stamps = np.sort(np_stamps)
    np_stamps = np_stamps - np_stamps[0]
    return np_stamps




import base64
from enum import Enum
import math
from pathlib import Path

import cv2
import numpy as np
import PIL
from PIL.Image import Image as PILImage
from shapely.geometry import Polygon

from pybsc.tile import squared_tile

try:
    from turbojpeg import TurboJPEG
    jpeg = TurboJPEG()
except Exception:
    jpeg = None


class ReturnType(Enum):
    BYTES = 0
    PILLOW = 1
    NDARRAY = 2


def str_to_pil_interpolation(interpolation):
    if interpolation == 'nearest':
        return PIL.Image.NEAREST
    elif interpolation == 'bilinear':
        return PIL.Image.BILINEAR
    elif interpolation == 'bicubic':
        return PIL.Image.BICUBIC
    elif interpolation == 'lanczos':
        return PIL.Image.LANCZOS
    else:
        raise ValueError(
            'Not valid Interpolation. '
            + 'Valid interpolation methods are '
            + 'nearest, bilinear, bicubic and lanczos.')


def pil_to_cv2_interpolation(interpolation):
    if isinstance(interpolation, str):
        interpolation = interpolation.lower()
        if interpolation == 'nearest':
            cv_interpolation = cv2.INTER_NEAREST
        elif interpolation == 'bilinear':
            cv_interpolation = cv2.INTER_LINEAR
        elif interpolation == 'bicubic':
            cv_interpolation = cv2.INTER_CUBIC
        elif interpolation == 'lanczos':
            cv_interpolation = cv2.INTER_LANCZOS4
        else:
            raise ValueError(
                'Not valid Interpolation. '
                + 'Valid interpolation methods are '
                + 'nearest, bilinear, bicubic and lanczos.')
    else:
        if interpolation == PIL.Image.NEAREST:
            cv_interpolation = cv2.INTER_NEAREST
        elif interpolation == PIL.Image.BILINEAR:
            cv_interpolation = cv2.INTER_LINEAR
        elif interpolation == PIL.Image.BICUBIC:
            cv_interpolation = cv2.INTER_CUBIC
        elif interpolation == PIL.Image.LANCZOS:
            cv_interpolation = cv2.INTER_LANCZOS4
        else:
            raise ValueError(
                'Not valid Interpolation. '
                + 'Valid interpolation methods are '
                + 'PIL.Image.NEAREST, PIL.Image.BILINEAR, '
                + 'PIL.Image.BICUBIC and PIL.Image.LANCZOS.')
    return cv_interpolation


def convert_to_numpy(input_data):
    """Convert the input data into a numpy.ndarray object.

    Parameters
    ----------
    input_data : str, Path, np.ndarray, or PIL.Image
        The input data to be converted. This can be a string or Path
        object representing a file path, a np.ndarray representing an image,
        or a PIL.Image object.

    Returns
    -------
    np.ndarray
        The converted np.ndarray object.

    Raises
    ------
    TypeError
        If the input_data is not of type str, Path, np.ndarray, or PIL.Image.
    """
    if isinstance(input_data, (str, Path)):
        return np.array(PIL.Image.open(str(input_data)))
    elif isinstance(input_data, np.ndarray):
        return input_data
    elif isinstance(input_data, PILImage.Image):
        return np.array(input_data)
    else:
        raise TypeError(
            "Invalid input type. Expected filepath, np.ndarray, or PIL.Image")


def convert_to_pil(input_data):
    """Convert the input data into a PIL.Image object.

    Parameters
    ----------
    input_data : str, Path, np.ndarray, or PIL.Image
        The input data to be converted.
        This can be a string or Path object representing a file path,
        a np.ndarray representing an image, or a PIL.Image object.

    Returns
    -------
    PIL.Image
        The converted PIL.Image object.

    Raises
    ------
    TypeError
        If the input_data is not of type str, Path, np.ndarray, or PIL.Image.
    """
    if isinstance(input_data, (str, Path)):
        return PIL.Image.open(str(input_data))
    elif isinstance(input_data, np.ndarray):
        return PIL.Image.fromarray(input_data)
    elif isinstance(input_data, PIL.Image):
        return input_data
    else:
        raise TypeError(
            "Invalid input type. Expected filepath, np.ndarray, or PIL.Image")


def _imread(image_path, color_type='bgr'):
    img = cv2.imread(str(image_path), cv2.IMREAD_UNCHANGED)

    if img is None:
        raise ValueError(f'Image not found at path: {image_path}')

    if color_type.lower() == 'gray':
        converted_img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    elif color_type.lower() == 'rgba':
        if img.shape[2] == 3:
            converted_img = cv2.cvtColor(img, cv2.COLOR_BGR2BGRA)
        elif img.shape[2] == 4:
            converted_img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)
    elif color_type.lower() == 'bgra':
        if img.shape[2] == 3:
            converted_img = cv2.cvtColor(img, cv2.COLOR_BGR2BGRA)
        elif img.shape[2] == 4:
            converted_img = img
    elif color_type.lower() == 'bgr':
        if img.shape[2] == 3:
            converted_img = img
        elif img.shape[2] == 4:
            converted_img = cv2.cvtColor(img, cv2.COLOR_BGRA2BGR)
    elif color_type.lower() == 'rgb':
        converted_img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    else:
        raise ValueError(f'Unsupported color type: {color_type}')
    return np.array(converted_img)


def imread(img_path, color_type='bgr',
           size_keeping_aspect_ratio_wrt_longside=None,
           clear_alpha=False):
    img = _imread(img_path, color_type=color_type)
    if size_keeping_aspect_ratio_wrt_longside:
        if color_type == 'bgra' or color_type == 'rgba':
            bgr = img[..., :3]
            alpha_img = img[..., 3]
            alpha_img = resize_keeping_aspect_ratio_wrt_longside(
                alpha_img, size_keeping_aspect_ratio_wrt_longside,
                interpolation='nearest')
            bgr = resize_keeping_aspect_ratio_wrt_longside(
                bgr, size_keeping_aspect_ratio_wrt_longside,
                interpolation='bilinear')
            img = np.dstack(
                [np.array(bgr, dtype=np.uint8),
                 np.array(alpha_img, dtype=np.uint8)[..., None]])
            if clear_alpha:
                img[..., 3] = 0
        else:
            img = resize_keeping_aspect_ratio_wrt_longside(
                img, size_keeping_aspect_ratio_wrt_longside,
                interpolation='bilinear')
    return img


def imwrite(image_path, img, color_type='bgr'):
    if color_type.lower() == 'gray':
        converted_img = PIL.Image.fromarray(img, 'L')
    elif color_type.lower() == 'rgba':
        converted_img = PIL.Image.fromarray(img, 'RGBA')
    elif color_type.lower() == 'bgra':
        b, g, r, a = img.split()
        converted_img = PIL.Image.merge("RGBA", (r, g, b, a))
    elif color_type.lower() == 'bgr':
        b, g, r = np.split(img, 3, axis=2)
        converted_img = PIL.Image.fromarray(
            np.concatenate((r, g, b), axis=2), 'RGB')
    elif color_type.lower() == 'rgb':
        converted_img = PIL.Image.fromarray(img, 'RGB')
    else:
        raise ValueError(f'Unsupported color type: {color_type}')

    converted_img.save(image_path)


def decode_image_cv2(b64encoded):
    bin = b64encoded.split(",")[-1]
    bin = base64.b64decode(bin)
    bin = np.frombuffer(bin, np.uint8)
    img = cv2.imdecode(bin, cv2.IMREAD_COLOR)
    return img


def decode_image_turbojpeg(b64encoded):
    bin = b64encoded.split(",")[-1]
    bin = base64.b64decode(bin)
    img = jpeg.decode(bin)
    return img


def decode_image(b64encoded):
    if jpeg is not None:
        img = decode_image_turbojpeg(b64encoded)
    else:
        img = decode_image_cv2(b64encoded)
    return img


def encode_image_turbojpeg(img):
    bin = jpeg.encode(img)
    b64encoded = base64.b64encode(bin).decode('ascii')
    return b64encoded


def encode_image_cv2(img, quality=90):
    encode_param = [int(cv2.IMWRITE_JPEG_QUALITY), quality]
    result, encimg = cv2.imencode('.jpg', img, encode_param)
    b64encoded = base64.b64encode(encimg).decode('ascii')
    return b64encoded


def encode_image(img):
    if jpeg is not None:
        img = encode_image_turbojpeg(img)
    else:
        img = encode_image_cv2(img)
    return img


def get_size(img):
    if isinstance(img, PILImage):
        return img.size
    elif isinstance(img, np.ndarray):
        return (img.shape[1], img.shape[0])
    else:
        raise RuntimeError(
            'input img should be PILImage or numpy.ndarray'
            + f', get {type(img)}')


def resize_keeping_aspect_ratio(img, width=None, height=None,
                                interpolation='bilinear',
                                return_scale=False):
    if (width and height) or (width is None and height is None):
        raise ValueError('Only width or height should be specified.')
    if isinstance(img, PILImage):
        if width == img.size[0] and height == img.size[1]:
            return img
        if width:
            scale = width / img.size[0]
            height = scale * img.size[1]
        else:
            scale = height / img.size[1]
            width = scale * img.size[0]
        height = int(height)
        width = int(width)
        resized_img = img.resize(
            (width, height),
            resample=str_to_pil_interpolation(interpolation))
    elif isinstance(img, np.ndarray):
        if width == img.shape[1] and height == img.shape[0]:
            return img
        if width:
            scale = width / img.shape[1]
            height = scale * img.shape[0]
        else:
            scale = height / img.shape[0]
            width = scale * img.shape[1]
        height = int(height)
        width = int(width)
        cv_interpolation = pil_to_cv2_interpolation(interpolation)
        resized_img = cv2.resize(img, (width, height),
                                 interpolation=cv_interpolation)
    else:
        raise ValueError(
            f"Input type {type(img)} is not supported.")
    if return_scale:
        return resized_img, scale
    else:
        return resized_img


def resize_keeping_aspect_ratio_wrt_longside(img, length,
                                             interpolation='bilinear',
                                             return_scale=False):
    if isinstance(img, PILImage):
        W, H = img.size
        aspect = W / H
        if H > W:
            width = length * aspect
            scale = length / H
            resized_img = img.resize(
                (int(width), int(length)),
                resample=str_to_pil_interpolation(interpolation))
        else:
            height = length / aspect
            scale = length / W
            resized_img = img.resize(
                (int(length), int(height)),
                resample=str_to_pil_interpolation(interpolation))
    elif isinstance(img, np.ndarray):
        cv_interpolation = pil_to_cv2_interpolation(interpolation)
        H, W = img.shape[:2]
        aspect = W / H
        if H > W:
            width = length * aspect
            scale = length / H
            resized_img = cv2.resize(
                img, (int(width), int(length)),
                interpolation=cv_interpolation)
        else:
            height = length / aspect
            scale = length / W
            resized_img = cv2.resize(
                img, (int(length), int(height)),
                interpolation=cv_interpolation)
    else:
        raise ValueError(
            f"Input type {type(img)} is not supported.")
    if return_scale:
        return resized_img, scale
    else:
        return resized_img


def _resize_keeping_aspect_ratio_wrt_longside_rgba(img, image_width):
    img = convert_to_numpy(img)
    try:
        img, mask = img[..., :3], img[..., 3]
    except IndexError:
        height = img.shape[0]
        width = img.shape[1]
        mask = 255 * np.ones(
            (height, width),
            dtype=np.uint8)
    pil_mask = PIL.Image.fromarray(mask)
    pil_mask = resize_keeping_aspect_ratio_wrt_longside(
        pil_mask, image_width,
        interpolation='nearest')
    pil_img = PIL.Image.fromarray(img)
    pil_img = resize_keeping_aspect_ratio_wrt_longside(
        pil_img, image_width,
        interpolation='bilinear')
    return np.concatenate(
        [np.array(pil_img, dtype=np.uint8),
         np.array(pil_mask, dtype=np.uint8).reshape(
             pil_mask.size[1],
             pil_mask.size[0],
             1)],
        axis=2)


def resize_keeping_aspect_ratio_wrt_target_size(
        img, width, height, interpolation='bilinear',
        background_color=(0, 0, 0)):
    if width == img.shape[1] and height == img.shape[0]:
        return img
    H, W, _ = img.shape
    ratio = min(float(height) / H, float(width) / W)
    M = np.array([[ratio, 0, 0],
                  [0, ratio, 0]], dtype=np.float32)
    dst = np.zeros((int(height), int(width), 3), dtype=img.dtype)
    return cv2.warpAffine(
        img, M,
        (int(width), int(height)),
        dst,
        cv2.INTER_CUBIC, cv2.BORDER_CONSTANT,
        background_color)


def squared_padding_image(img, length=None,
                          return_offset=None):
    H, W = img.shape[:2]
    if H > W:
        if length is not None:
            img = resize_keeping_aspect_ratio_wrt_longside(img, length)
        margin = img.shape[0] - img.shape[1]
        offset = (margin // 2, 0)
        img = np.pad(img,
                     [(0, 0),
                      (margin // 2, margin - margin // 2),
                      (0, 0)], 'constant')
    else:
        if length is not None:
            img = resize_keeping_aspect_ratio_wrt_longside(img, length)
        margin = img.shape[1] - img.shape[0]
        offset = (0, margin // 2)
        img = np.pad(img,
                     [(margin // 2, margin - margin // 2),
                      (0, 0), (0, 0)], 'constant')
    if return_offset:
        return img, offset
    else:
        return img


def concat_with_keeping_aspect(
        imgs, width, height,
        tile_shape=None):
    if len(imgs) == 0:
        raise ValueError
    if tile_shape is None:
        tile_x, tile_y = squared_tile(len(imgs))
    else:
        tile_x, tile_y = tile_shape

    w = width // tile_x
    h = height // tile_y

    ret = []
    max_height = h
    max_width = w
    for img in imgs:
        if img.shape[1] / w > img.shape[0] / h:
            tmp_img = resize_keeping_aspect_ratio(img, width=w, height=None)
        else:
            tmp_img = resize_keeping_aspect_ratio(img, width=None, height=h)
        ret.append(tmp_img)

    canvas = np.zeros((height, width, 3),
                      dtype=np.uint8)

    i = 0
    for y in range(tile_y):
        for x in range(tile_x):
            lh = (max_height - ret[i].shape[0]) // 2
            rh = (max_height - ret[i].shape[0]) - lh
            lw = (max_width - ret[i].shape[1]) // 2
            rw = (max_width - ret[i].shape[1]) - lw
            img = np.pad(ret[i],
                         [(lh, rh),
                          (lw, rw),
                          (0, 0)], 'constant')
            canvas[y * max_height:(y + 1) * max_height,
                   x * max_width:(x + 1) * max_width] = img
            i += 1
            if i >= len(imgs):
                break
        if i >= len(imgs):
            break
    return canvas


def mask_to_bbox(mask, threshold=0):
    if isinstance(mask, PILImage):
        mask = np.array(mask)
    elif isinstance(mask, np.ndarray):
        pass
    else:
        raise TypeError(f'Invalid input image type, {type(mask)}')
    mask = mask > threshold
    mask_indexes = np.where(mask)
    y_min = np.min(mask_indexes[0])
    y_max = np.max(mask_indexes[0])
    x_min = np.min(mask_indexes[1])
    x_max = np.max(mask_indexes[1])
    return (y_min, x_min, y_max, x_max)


def masks_to_bboxes(mask):
    R, _, _ = mask.shape
    instance_index, ys, xs = np.nonzero(mask)
    bboxes = np.zeros((R, 4), dtype=np.float32)
    for i in range(R):
        ys_i = ys[instance_index == i]
        xs_i = xs[instance_index == i]
        if len(ys_i) == 0:
            continue
        y_min = ys_i.min()
        x_min = xs_i.min()
        y_max = ys_i.max() + 1
        x_max = xs_i.max() + 1
        bboxes[i] = np.array(
            [x_min, y_min, x_max, y_max],
            dtype=np.float32)
    return bboxes


def alpha_blend(a_img, b_img, alpha=0.5):
    viz = cv2.addWeighted(a_img, alpha, b_img, 1 - alpha, 0)
    return viz


def zoom(img, ratio=1.0, interpolation='bilinear'):
    """zoom function resize and crop images.

    Parameters
    ----------
    img : np.ndarray
        input image (C, H, W)
    ration : float
        zoom ratio
        should be greater than 1.0.
    Returns
    -------
    cropped_img : np.ndarray
        zoomed image
    """
    if ratio < 1.0:
        raise ValueError(f'ratio should be greater than 1.0, but given {ratio}')
    w, h = get_size(img)
    H = int(h * ratio)
    W = int(w * ratio)
    resized_img = resize_keeping_aspect_ratio(
        img, height=H,
        interpolation=interpolation)
    cropped_img = resized_img[(H - h) // 2:(H - h) // 2 + h,
                              (W - w) // 2:(W - w) // 2 + w,
                              :]
    return cropped_img


def tile_image(wh_size, tile_size_wh, window_size=None):
    w, h = wh_size
    if isinstance(tile_size_wh, tuple) or isinstance(tile_size_wh, list):
        tile_width, tile_height = tile_size_wh
    else:
        tile_width, tile_height = tile_size_wh, tile_size_wh

    if window_size is None:
        window_h = tile_height
        window_w = tile_width
    else:
        if isinstance(window_size, tuple) or isinstance(window_size, list):
            window_w, window_h = window_size
        else:
            window_w = window_size
            window_h = window_size

    tile_w = tile_width
    tile_h = tile_height

    x = np.arange(0, w - tile_w, window_w)
    if (w - tile_w) % tile_w != 0:
        x = np.concatenate([x, np.array([w - tile_w])])
    y = np.arange(0, h - tile_h, window_h)
    if (h - tile_h) % tile_h != 0:
        y = np.concatenate([y, np.array([h - tile_h])])

    sx, sy = np.meshgrid(x, y)
    sx = sx.reshape(-1)
    ex = sx + tile_w
    sy = sy.reshape(-1)
    ey = sy + tile_h

    return np.array(
        [((sx[i], sy[i], ex[i] - sx[i], ey[i] - sy[i]))
         for i in range(len(sx))])


def create_tile_image(image_list,
                      image_size=300,
                      num_tiles_per_row=10,
                      background_color=(255, 255, 255)):
    """Create a tiled image from a list of images.

    Parameters
    ----------
    image_list : list
        A list of images to be tiled. Each image can be a filepath
        (str or Path), np.ndarray, or PIL.Image.
    image_size : int
        The size for each individual image tile. Defaults to 300.
    num_tiles_per_row : int, optional
        The number of image tiles per row in the output image.
        Defaults to 10.
    background_color : tuple, optional
        The background color for the output image in (R, G, B).
        Defaults to (255, 255, 255).

    Returns
    -------
    PIL.Image
        The created tiled image.

    Raises
    ------
    ValueError
        If any of the images in image_list cannot be opened or read.
    """
    num_images = len(image_list)
    num_rows = math.ceil(num_images / num_tiles_per_row)

    tile_width = image_size * num_tiles_per_row
    tile_height = image_size * num_rows
    tile_image = PIL.Image.new(
        'RGBA', (tile_width, tile_height),
        tuple(list(background_color) + [0]))

    for i, image in enumerate(image_list):
        tile_x = (i % num_tiles_per_row) * image_size
        tile_y = (i // num_tiles_per_row) * image_size
        pil_img = convert_to_pil(
            _resize_keeping_aspect_ratio_wrt_longside_rgba(
                image, image_size))
        tile_image.paste(pil_img, (tile_x, tile_y))
    return tile_image


def get_bboxes_from_tile(bboxes, tile):
    new_bboxes = []
    for y1, x1, y2, x2 in bboxes:
        new_bboxes.append(Polygon([(x1, y1), (x2, y1), (x2, y2), (x1, y2)]))
    bboxes = new_bboxes
    x1, y1, w, h = tile
    x2, y2 = x1 + w, y1 + h
    pol = Polygon([(x1, y1), (x2, y1), (x2, y2), (x1, y2)])
    sliced_bboxes = []
    indices = []
    for box_idx, box in enumerate(bboxes):
        if pol.intersects(box):
            inter = pol.intersection(box)
            # get the smallest polygon
            # (with sides parallel to the coordinate axes)
            # that contains the intersection
            new_box = inter.envelope

            # get coordinates of polygon vertices
            x, y = new_box.exterior.coords.xy
            x = x - x1
            y = y - y1

            sliced_bboxes.append([min(y), min(x), max(y), max(x)])
            indices.append(box_idx)
    return sliced_bboxes, indices


def non_maximum_suppression(bbox, thresh, score=None, limit=None):
    if len(bbox) == 0:
        return np.zeros((0,), dtype=np.int32)

    if score is not None:
        order = score.argsort()[::-1]
        bbox = bbox[order]
    bbox_area = np.prod(bbox[:, 2:] - bbox[:, :2], axis=1)

    indices = np.zeros(bbox.shape[0], dtype=bool)
    for i, b in enumerate(bbox):
        tl = np.maximum(b[:2], bbox[indices, :2])
        br = np.minimum(b[2:], bbox[indices, 2:])
        area = np.prod(br - tl, axis=1) * (tl < br).all(axis=1)

        iou = area / (bbox_area[i] + bbox_area[indices] - area)
        if (iou >= thresh).any():
            continue

        indices[i] = True
        if limit is not None and np.count_nonzero(indices) >= limit:
            break

    indices = np.where(indices)[0]
    if score is not None:
        indices = order[indices]
    return indices.astype(np.int32)


def rotate(pil_img, mask=None, angle=360):
    return_np = False
    concat = False
    if isinstance(pil_img, np.ndarray):
        if pil_img.shape[2] == 4:
            concat = True
            pil_img, mask = PIL.Image.fromarray(pil_img[..., :3]), \
                PIL.Image.fromarray(pil_img[..., 3])
        else:
            pil_img = PIL.Image.fromarray(pil_img)
            if mask is not None:
                mask = PIL.Image.fromarray(mask)
        return_np = True

    rot_img = pil_img.rotate(angle, resample=PIL.Image.BILINEAR,
                             expand=True)
    if mask is not None:
        if return_np:
            if concat:
                rgb = np.array(rot_img)
                mask = np.array(mask.rotate(
                    angle,
                    resample=PIL.Image.NEAREST,
                    expand=True), dtype=np.uint8)
                return np.concatenate(
                    [rgb, mask[..., None]], axis=2)
            return np.array(rot_img), np.array(mask.rotate(
                angle,
                resample=PIL.Image.NEAREST,
                expand=True), dtype=np.uint8)
        else:
            return rot_img, mask.rotate(
                angle,
                resample=PIL.Image.NEAREST,
                expand=True)
    if return_np:
        return np.array(rot_img)
    return rot_img


def rescale(pil_img, mask=None, scale=1.0):
    if isinstance(pil_img, np.ndarray):
        concat = False
        if pil_img.shape[2] == 4:
            concat = True
            pil_img, mask = PIL.Image.fromarray(
                pil_img[..., :3]), \
                PIL.Image.fromarray(pil_img[..., 3])
        else:
            pil_img = PIL.Image.fromarray(pil_img)
            if mask is not None:
                mask = PIL.Image.fromarray(mask)
        w, h = pil_img.size
        w = int(np.ceil(scale * pil_img.size[0]))
        h = int(np.ceil(scale * pil_img.size[1]))
        pil_img = pil_img.resize((w, h), PIL.Image.BILINEAR)
        rgb = np.array(pil_img, dtype=np.uint8)
        if mask is not None:
            mask = mask.resize((w, h), PIL.Image.NEAREST)
            mask = np.array(mask, dtype=np.uint8)
            if concat:
                np.concatenate(
                    [rgb, mask[..., None]], axis=2)
            else:
                return rgb, mask
        return rgb

    else:
        w, h = pil_img.size
        w = int(np.ceil(scale * pil_img.size[0]))
        h = int(np.ceil(scale * pil_img.size[1]))
        pil_img = pil_img.resize((w, h), PIL.Image.BILINEAR)
        if mask is not None:
            mask = mask.resize((w, h), PIL.Image.NEAREST)
            return pil_img, mask
        return pil_img


def apply_mask(bgr_img, mask, fill_value=255, alpha=255,
               use_alpha=False):
    """Apply a mask to an image.

    This function applies a mask to an image, with an optional fill value
    for unmasked areas. If the fill value is a list or tuple, it's assumed
    to represent RGB values.

    Parameters
    ----------
    bgr_img : ndarray
        An input image in BGR format.
    mask : ndarray
        The mask to apply. Any areas where the mask is 0 will be filled
        with the fill_value.
    fill_value : int or list/tuple of int, optional
        The value to fill unmasked areas with. If a list or tuple, it should
        represent BGR values. Defaults to 255.
    alpha : int, optional
        Not used in this function. Defaults to 255.

    Returns
    -------
    img : ndarray
        The image with the mask applied.
    """
    img = bgr_img.copy()
    if len(img.shape) == 3:
        img = img[..., :3]
    if isinstance(fill_value, (list, tuple)):
        fill_value = np.array(fill_value, dtype=np.uint8)
    else:
        fill_value = np.array(
            [fill_value, fill_value, fill_value],
            dtype=np.uint8)
    img[mask == 0] = fill_value
    if use_alpha:
        if len(img.shape) == 2:
            img = cv2.cvtColor(img, cv2.GRAY2BGR)
        return np.dstack([img, mask])
    return img


def add_alpha_channel(img, alpha=255):
    """Add an alpha channel to an image.

    If the image already has an alpha channel, the image is returned as is.
    If the image has three channels (BGR), a fourth alpha channel is added.

    Parameters
    ----------
    img : ndarray
        The input image. Should have either 3 (BGR) or 4 (BGRA) channels.
    alpha : int, optional
        The alpha value to set for the new channel. Defaults to 255.

    Returns
    -------
    img_rgba : ndarray
        The image with an added alpha channel.

    Raises
    ------
    ValueError
        If the input image does not have either 3 (BGR) or 4 (BGRA) channels.
    """
    # Checking if the image has 3 channels (Red, Green, Blue)
    if img.shape[2] == 3:
        h, w, _ = img.shape
        # Create a new image with 4 channels
        img_rgba = np.zeros((h, w, 4), dtype=img.dtype)
        img_rgba[:, :, :3] = img
        img_rgba[:, :, 3] = alpha
        return img_rgba
    elif img.shape[2] == 4:  # If the image already has an alpha channel
        return img
    else:
        raise ValueError(
            "Invalid number of channels. "
            + "The image should either have 3 (BGR) or 4 (BGRA) channels.")



import json
import os.path as osp

import numpy as np


class NumpyArrayEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyArrayEncoder, self).default(obj)


def load_json(file_path, backend='builtin'):
    """Load json function.

    Parameters
    ----------
    file_path : str or pathlib.PosixPath
        json file path

    Returns
    -------
    data : dict
        loaded json data
    """
    if not osp.exists(str(file_path)):
        raise OSError(f'{file_path!s} not exists')
    with open(str(file_path)) as f:
        if backend == 'builtin':
            try:
                return json.load(f)
            except Exception as e:
                raise OSError(f'Error {e!s}. Not valid json file {file_path!s}')
        elif backend == 'orjson':
            import orjson
            try:
                return orjson.loads(f.read())
            except Exception as e:
                raise OSError(f'Error {e!s}. Not valid json file {file_path!s}')
        else:
            raise NotImplementedError(
                f"Not supported backend {backend}")


def save_json(data, filename,
              save_pretty=True,
              sort_keys=True,
              ensure_ascii=True,
              save_as_jsonlines=False,
              backend='builtin'):
    """Save json function.

    Parameters
    ----------
    data : dict or list[dict]
        save data
    filename : str
        save path
    """
    filename = str(filename)
    if save_as_jsonlines:
        import jsonlines
        with jsonlines.open(filename, mode='w') as writer:
            if isinstance(data, list):
                for d in data:
                    writer.write(d)
            else:
                writer.write(data)
    elif backend == 'builtin':
        with open(filename, "w") as f:
            if save_pretty:
                f.write(json.dumps(data, indent=4,
                                   ensure_ascii=ensure_ascii,
                                   sort_keys=sort_keys,
                                   separators=(',', ': '),
                                   cls=NumpyArrayEncoder))
            else:
                json.dump(data, f)
            f.write('\n')
    elif backend == 'orjson':
        import orjson
        with open(filename, "wb") as f:
            if save_pretty:
                f.write(orjson.dumps(
                    data,
                    option=orjson.OPT_SORT_KEYS | orjson.OPT_INDENT_2))
            else:
                f.write(orjson.dumps(data))
            f.write(b'\n')
    else:
        raise NotImplementedError(f"Not supported backend {backend}")



import os

from pybsc.pycompat import PY2


def makedirs(name, mode=0o777, exist_ok=True):
    """An wrapper of os.makedirs that accepts exist_ok.

    Parameters
    ----------
    name : str
        path of directory
    exist_ok : bool
        if True, accepts the existence of the directory.

    Examples
    --------
    >>> from pybsc import makedirs
    >>> makedirs('/tmp/result_directory')
    """
    name = str(name)
    if PY2:
        try:
            os.makedirs(name, mode)
        except OSError:
            if not (exist_ok and os.path.isdir(name)):
                raise OSError(
                    f'Directory {name} already exists. '
                    + 'Set exist_ok = True if the directory can exist.'
                    )
    else:
        os.makedirs(name, mode, exist_ok=exist_ok)



import os.path as osp
import platform
import subprocess


def _open(filepath):
    pf = platform.system()
    if pf == 'Windows':
        opener = 'start'
    elif pf == 'Darwin':
        opener = 'open'
    else:
        opener = 'xdg-open'
    return f'{opener} {filepath}'


def opener(filepath):
    if not osp.exists(filepath):
        raise OSError(f'{filepath} not exists')
    proc = subprocess.Popen(
        _open(filepath),
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        shell=True
    )
    proc.wait()
    return proc



import os.path as osp

import yaml


def load_yaml(file_path, Loader=yaml.SafeLoader):
    """Load a YAML file into a Python dict.

    Parameters
    ----------
    file_path : str or pathlib.PosixPath
        The path to the YAML file.

    Returns
    -------
    data : dict
        A dict with the loaded yaml data.
    """
    if not osp.exists(str(file_path)):
        raise OSError(f'{file_path!s} not exists')
    with open(osp.expanduser(file_path)) as f:
        data = yaml.load(f, Loader=Loader)
    return data


def save_yaml(data, file_path):
    """Save a dict to a YAML file.

    Parameters
    ----------
    data : dict
        A dict with the loaded yaml data.
    file_path : str or pathlib.PosixPath
        The path to the YAML file.
    """
    with open(osp.expanduser(file_path), 'w') as f:
        yaml.dump(data, f, default_flow_style=None)





def nsplit(xlst, n):
    """Split input list into n.

    Parameters
    ----------
    xlst : list[object]
        input list
    n : int
       number of divide.

    Returns
    -------
    ret : list[list[object]]
        n list.

    Examples
    --------
    >>> from pybsc import nsplit
    >>> nsplit([1, 2, 3], 3)
    [[1], [2], [3]]
    >>> nsplit([1, 2, 3], 2)
    [[1, 2], [3]]
    """
    total_n = len(xlst)
    d = int((total_n + n - 1) / n)
    i = 0
    ret = []
    while i < total_n:
        ret.append(xlst[i:i + d])
        i += d
    return ret



import pickle


def load_pickle(file_path):
    """Load pickle data.

    Parameters
    ----------
    file_path : str or pathlib.PosixPath
        pickle file path
    """
    with open(file_path, 'rb') as f:
        data = pickle.load(f)
    return data



import cv2
import numpy as np


class ColorFilter:

    def __init__(self):
        pass


class HSVColorFilter(ColorFilter):

    def __init__(self,
                 h_max=360, h_min=0,
                 s_max=256, s_min=0,
                 v_max=256, v_min=0):
        super(HSVColorFilter, self).__init__()
        self.update_condition(
            h_max, h_min, s_max, s_min, v_max, v_min)

    def update_condition(
            self, h_max=None, h_min=None,
            s_max=None, s_min=None, v_max=None, v_min=None):
        if h_min is not None:
            self.h_min = h_min
        if h_max is not None:
            self.h_max = h_max
        if s_min is not None:
            self.s_min = s_min
        if s_max is not None:
            self.s_max = s_max
        if v_min is not None:
            self.v_min = v_min
        if v_max is not None:
            self.v_max = v_max
        if self.s_max < self.s_min:
            self.s_max, self.s_min = self.s_min, self.s_max
        if self.v_max < self.v_min:
            self.v_max, self.v_min = self.v_min, self.v_max
        self.lower_color_range = np.array([
            self.h_min / 2.0, self.s_min, self.v_min, 0])
        self.upper_color_range = np.array([
            self.h_max / 2.0, self.s_max, self.v_max, 0])

    def filter(self, bgr_img):
        hsv_img = cv2.cvtColor(bgr_img, cv2.COLOR_BGR2HSV)
        if self.lower_color_range[0] < self.upper_color_range[0]:
            mask = cv2.inRange(hsv_img,
                               self.lower_color_range,
                               self.upper_color_range)
        else:
            lower_color_range_0 = np.array((0, self.s_min, self.v_min, 0),
                                           dtype=np.float64)
            upper_color_range_0 = np.array((self.h_max / 2.0,
                                            self.s_max, self.v_max, 0),
                                           dtype=np.float64)
            lower_color_range_360 = np.array((self.h_min / 2.0,
                                              self.s_min, self.v_min, 0),
                                             dtype=np.float64)
            upper_color_range_360 = np.array(
                (360 / 2.0, self.s_max, self.v_max, 0), dtype=np.float64)
            mask_0 = cv2.inRange(hsv_img,
                                 lower_color_range_0, upper_color_range_0)
            mask_360 = cv2.inRange(
                hsv_img, lower_color_range_360, upper_color_range_360)
            mask = np.array(
                np.logical_or(mask_0, mask_360) * 255, dtype=np.uint8)
        return mask


class LabColorFilter(ColorFilter):

    def __init__(self,
                 l_max=255,
                 l_min=0,
                 a_max=255,
                 a_min=0,
                 b_max=255,
                 b_min=0):
        super(LabColorFilter, self).__init__()
        self.update_condition(
            l_max, l_min,
            a_max, a_min,
            b_max, b_min)

    def update_condition(self,
                         l_max=None, l_min=None,
                         a_max=None, a_min=None,
                         b_max=None, b_min=None):
        if l_min is not None:
            self.l_min = l_min
        if l_max is not None:
            self.l_max = l_max
        if a_min is not None:
            self.a_min = a_min
        if a_max is not None:
            self.a_max = a_max
        if b_min is not None:
            self.b_min = b_min
        if b_max is not None:
            self.b_max = b_max
        self.lower_color_range = np.array(
            [self.l_min, self.a_min, self.b_min])
        self.upper_color_range = np.array(
            [self.l_max, self.a_max, self.b_max])

    def filter(self, bgr_img):
        lab_img = cv2.cvtColor(bgr_img, cv2.COLOR_BGR2Lab)
        print(self.lower_color_range)
        print(self.upper_color_range)
        mask = cv2.inRange(lab_img,
                           self.lower_color_range,
                           self.upper_color_range)
        return mask


def hsv_color_filter(bgr_img,
                     h_max=360, h_min=0,
                     s_max=256, s_min=0,
                     v_max=256, v_min=0):
    return HSVColorFilter(
        h_max, h_min, s_max, s_min, v_max, v_min).filter(bgr_img)


def lab_color_filter(bgr_img,
                     l_max=255, l_min=0,
                     a_max=255, a_min=0,
                     b_max=255, b_min=0):
    return LabColorFilter(
        l_max, l_min, a_max, a_min, b_max, b_min).filter(bgr_img)



from .pycompat import PY3

if PY3:
    from shutil import which as _which
else:
    from distutils.spawn import find_executable
    import os

    def _which(command):
        _search_path = os.environ['PATH']
        _executable = None
        for _name in [command]:
            _executable = find_executable(
                _name, path=_search_path)
            if _executable is not None:
                break
        return _executable


def which(cmd):
    return _which(cmd)



import logging

import coloredlogs

base_logger = logging.getLogger(__name__)


def deprecated_warning(msg, cls=UserWarning,
                       logger=None):
    if logger is None:
        logger = base_logger
    coloredlogs.install(level='DEBUG', logger=logger)

    def _deprecated_warning(fn):
        def decorated(*args, **kwargs):
            logger.warning(msg)
            return fn(*args, **kwargs)

        return decorated

    return _deprecated_warning



import numpy as np
import scipy
import scipy.cluster


def pillow_image_to_simple_bitmap(pillow_image, mask=None):
    if mask is not None:
        bitmap = np.array(pillow_image, dtype=np.int16)[np.array(mask) > 150]
    else:
        bitmap = np.array(pillow_image, dtype=np.int16)
        shape = bitmap.shape
        bitmap = bitmap.reshape(scipy.product(shape[:2]), shape[2])
    bitmap = bitmap.astype(np.float)
    return bitmap


def top_n_colors(pillow_image, top_n, num_of_clusters, mask=None):
    clustering = scipy.cluster.vq.kmeans
    bitmap = pillow_image_to_simple_bitmap(pillow_image, mask)
    clusters, _ = clustering(bitmap, num_of_clusters)
    quntized, _ = scipy.cluster.vq.vq(bitmap, clusters)
    histgrams, _ = scipy.histogram(quntized, len(clusters))
    order = np.argsort(histgrams)[::-1][:top_n]
    for idx in range(min(top_n, len(order))):
        rgb = clusters.astype(int)[order[idx]].tolist()
        yield rgb



from pathlib import Path
import shutil
from tempfile import NamedTemporaryFile

from fastapi import UploadFile


def save_upload_file_to_tmp(upload_file: UploadFile) -> Path:
    try:
        suffix = Path(upload_file.filename).suffix
        with NamedTemporaryFile(delete=False, suffix=suffix) as tmp:
            shutil.copyfileobj(upload_file.file, tmp)
        tmp_path = Path(tmp.name)
    finally:
        upload_file.file.close()
    return tmp_path



import contextlib
import datetime as datetime_module
import sys

time_isoformat = datetime_module.time.isoformat


time_fromisoformat = datetime_module.time.fromisoformat


def timedelta_isoformat(timedelta, timespec='microseconds'):
    time = (datetime_module.datetime.min + timedelta).time()
    return time_isoformat(time, timespec)


@contextlib.contextmanager
def measure(log_name='Total elapsed time',
            file=sys.stderr):
    """Measure time with contextlib.

    Parameters
    ----------
    log_name : str
        log name
    file : writable object
        writable object such as sys.stderr.

    Examples
    --------
    >>> from pybsc import measure
    >>> with measure('measure example'):
    ...     time.sleep(2.0)
    measure example: 00:00:02.002104
    """
    start_time = datetime_module.datetime.now()
    yield
    duration = datetime_module.datetime.now() - start_time
    now_string = timedelta_isoformat(duration)
    file.write('{log_name}: {now_string}\n'.format(
        **locals()))



# flake8: noqa

from .generate_texture import generate_check_texture
from .generate_texture import generate_gradation_texture
from .generate_texture import generate_single_color_texture
from .generate_texture import generate_synthetic_texture



import cv2
import numpy as np


def estimate_blur(img):
    if img.ndim == 3:
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    blur_map = cv2.Laplacian(img, cv2.CV_64F)
    score = np.var(blur_map)
    return blur_map, score


def calculate_blur_map(blur_map, blur_sigma=5,
                       median_blur_sigma=5, min_abs=0.5):
    abs_image = np.abs(blur_map).astype(np.float32)
    abs_image[abs_image < min_abs] = min_abs
    abs_image = np.log(abs_image)
    abs_image = cv2.blur(abs_image, (blur_sigma, blur_sigma))
    abs_image = np.array(abs_image, dtype=np.uint8)
    return cv2.medianBlur(abs_image, median_blur_sigma)



import numpy as np


def generate_gaussian_heatmap(shape, xy, sigma=1.0):
    """Return generated gaussian heatmap.

    Parameters
    ==========
    shape : tuple(int, int)
        width, height
    xy : tuple(int, int)
        x y position.
    sigma : float
        sigma
    """
    width, height = shape
    x, y = xy
    grid_x = np.tile(np.arange(width), (height, 1))
    grid_y = np.tile(np.arange(height), (width, 1)).transpose()
    grid_distance = (grid_x - x) ** 2 + (grid_y - y) ** 2
    gaussian_heatmap = np.exp(-0.5 * grid_distance / sigma**2)
    return gaussian_heatmap



def invert_dict(input_dict):
    """Inverrt key-value relationship in a dictionary

    Parameters
    ----------
    input_dict : dict
        input dictionary.
    """
    output_dict = {}
    for key, value in input_dict.items():
        output_dict[value] = key
    return output_dict



import os
import sys

import pyperclip


class ClipboardOutput:

    def __init__(self):
        self.buf = ''

    def write(self, arg):
        self.buf += arg


class stdout_to_clipboard:

    def __init__(self, print_stdout=True):
        sys.stdout.flush()
        self._org_stdout = sys.stdout
        self._old_stdout_fno = os.dup(sys.stdout.fileno())
        self._clip = ClipboardOutput()
        self.print_stdout = print_stdout

    def __enter__(self):
        sys.stdout = self._clip

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self._org_stdout
        sys.stdout.flush()
        os.dup2(self._old_stdout_fno, 1)
        pyperclip.copy(self._clip.buf)
        if self.print_stdout:
            print(self._clip.buf)



import sys


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via input() and return their answer.

    Parameters
    ----------
    question : str
        The question presented to the user.
    default : str, optional
        The presumed answer if the user just hits <Enter>.
        It must be one of "yes", "no", or None. Defaults to "yes".

    Returns
    -------
    bool
        True for "yes" or False for "no".

    Raises
    ------
    ValueError
        If the default answer is not one of "yes", "no", or None.

    Notes
    -----
    The function repeatedly prompts the user until they provide a valid
    yes/no response. The response is case-insensitive and accepts
    'y'/'yes' or 'n'/'no'.

    Examples
    --------
    >>> query_yes_no("Do you like Python?")
    Do you like Python? [Y/n]
    # User inputs 'yes'
    True

    >>> query_yes_no("Do you like Python?", default=None)
    Do you like Python? [y/n]
    # User inputs 'n'
    False
    """

    valid = {"yes": True, "y": True, "no": False, "n": False}
    prompt_options = {"yes": " [Y/n] ", "no": " [y/N] ", None: " [y/n] "}

    if default not in prompt_options:
        raise ValueError(f"Invalid default answer: '{default}'")

    prompt = prompt_options[default]

    while True:
        sys.stdout.write(f"{question}{prompt}")
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write(
                "Please respond with 'yes' or 'no' (or 'y' or 'n').\n")




import os
import shlex
import subprocess
import sys

from setuptools import find_packages
from setuptools import setup

version = "0.1.54"


if sys.argv[-1] == "release":
    # Release via github-actions.
    commands = [
        f'git tag v{version:s}',
        'git push origin master --tag',
    ]
    for cmd in commands:
        print(f'+ {cmd}')
        subprocess.check_call(shlex.split(cmd))
    sys.exit(0)


setup_requires = []

with open('requirements.txt') as f:
    install_requires = []
    for line in f:
        req = line.split('#')[0].strip()
        install_requires.append(req)

image_install_requires = []
with open('requirements_image.txt') as f:
    image_install_requires = []
    for line in f:
        req = line.split('#')[0].strip()
        image_install_requires.append(req)

video_install_requires = []
with open('requirements_video.txt') as f:
    video_install_requires = []
    for line in f:
        req = line.split('#')[0].strip()
        video_install_requires.append(req)

tts_install_requires = []
with open('requirements_tts.txt') as f:
    tts_install_requires = []
    for line in f:
        req = line.split('#')[0].strip()
        tts_install_requires.append(req)


def listup_package_data():
    data_files = []
    for root, _, files in os.walk('pybsc/data'):
        for filename in files:
            data_files.append(
                os.path.join(
                    root[len('pybsc/'):],
                    filename))
    return data_files


setup(
    name="pybsc",
    version=version,
    description="A little useful library (like a Backscratcher).",
    author="iory",
    author_email="ab.ioryz@gmail.com",
    url="https://github.com/iory/pybsc",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    packages=find_packages(),
    package_data={'pybsc': listup_package_data()},
    zip_safe=False,
    setup_requires=setup_requires,
    install_requires=install_requires,
    extras_require={
        'image': image_install_requires,
        'video': video_install_requires,
        'tts': tts_install_requires,
        'all': (image_install_requires
                + video_install_requires
                + tts_install_requires),
    },
    entry_points={
        "console_scripts": [
            "readable-json=pybsc.apps.readable_json:main",
        ]
    },
)



from collections import defaultdict
import os
import os.path as osp


def get_target_ext(base_path, exts):
    exts = tuple(exts)
    base_path = str(base_path)
    exts_dict = defaultdict(int)
    for ext in exts:
        exts_dict[ext] = 1
    for dirpath, _, files in os.walk(base_path):
        for f in files:
            _, ext = osp.splitext(f.lower())
            if exts_dict[ext]:
                yield os.path.join(dirpath, f)



import string

printable = set(string.printable)


def removesuffix(s, suffix):
    return s[: -len(suffix)] if s.endswith(suffix) else s


def remove_non_ascii(s):
    return ''.join(filter(lambda x: x in printable, s))



import os.path as osp
import time

import paramiko

from pybsc import run_command


def return_true():
    return True


def create_tunnel_through_bastion(
        target_host, target_user,
        key_filepath=None,
        bastion_host=None, bastion_user=None):
    if key_filepath is None:
        key_filepath = osp.expanduser(osp.join('~', '.ssh', 'id_rsa'))
    target = paramiko.SSHClient()
    target.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    if bastion_host and bastion_user:
        bastion = paramiko.SSHClient()
        bastion.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        bastion.connect(bastion_host, username=bastion_user,
                        key_filename=key_filepath)

        bastion_transport = bastion.get_transport()
        dest_addr = (target_host, 22)  # Destination server IP and SSH port
        local_addr = ('127.0.0.1', 22)  # Local server IP and SSH port
        channel = bastion_transport.open_channel(
            "direct-tcpip", dest_addr, local_addr)

        target.connect(target_host, username=target_user,
                       key_filename=key_filepath, sock=channel)
    else:
        target.connect(target_host, username=target_user,
                       key_filename=key_filepath)
    return target


class SSHExecutor:

    def __init__(
            self, target_host, target_user,
            key_filepath=None,
            bastion_host=None, bastion_user=None):
        if key_filepath is None:
            key_filepath = osp.expanduser(osp.join('~', '.ssh', 'id_rsa'))
        self._target_host = target_host
        self._target_user = target_user
        self._key_filepath = key_filepath
        self._bastion_host = bastion_host
        self._bastion_user = bastion_user
        self._client = create_tunnel_through_bastion(
            target_host, target_user, key_filepath, bastion_host, bastion_user)

    def __del__(self):
        self._client.close()

    def execute_command(self, command):
        stdin, stdout, stderr = self._client.exec_command(command)
        output = stdout.read()
        return output

    def watch_dog(self, remote_file_path, poll_interval=5, condition=None):
        if condition is None:
            condition = return_true
        while condition():
            try:
                stdin, stdout, stderr = self._client.exec_command(
                    f"test -f {remote_file_path} && echo exists || echo not_exists")
                output = stdout.read().decode().strip()
                if "exists" == output:
                    print(f"File {remote_file_path} exists on remote server.")
                    break
                else:
                    print(f"Waiting for file {remote_file_path} to be created on remote server.")
                    time.sleep(poll_interval)
            except Exception as e:
                print(f"Error in watch_dog: {e}")
                break
        return condition()

    def execute_command_tmux(self, command, session_name):
        self.execute_command(f"tmux new-session -d -s {session_name}")
        tmux_command = f"tmux send-keys -t {session_name} '{command}' Enter"
        self.execute_command(tmux_command)

    def kill_tmux_session(self, session_name):
        self.execute_command(f'tmux kill-session -t {session_name}')

    def rsync(self, local_path, remote_path, is_upload=True):
        ssh_options = ["ssh", f"-i {self._key_filepath}",
                       '-o', 'StrictHostKeyChecking=no']

        if self._bastion_host is not None and self._bastion_user is not None:
            proxy_command = (
                f"-o ProxyCommand='ssh -o StrictHostKeyChecking=no -W %h:%p -i {self._key_filepath} "  # NOQA
                f"{self._bastion_user}@{self._bastion_host}'"
            )
            ssh_options.append(proxy_command)

        ssh_option_str = ' '.join(ssh_options)
        rsync_command = f"rsync -avz -e \"{ssh_option_str}\""

        if is_upload and osp.isdir(local_path):
            rsync_command += " -r"

        remote = f"{self._target_user}@{self._target_host}:{remote_path}"

        if is_upload:
            full_command = f"{rsync_command} \"{local_path}\" \"{remote}\""
        else:
            full_command = f"{rsync_command} \"{remote}\" \"{local_path}\""
        return run_command(full_command, shell=True)



import os

from azure.cognitiveservices.speech import SpeechConfig
from azure.cognitiveservices.speech import SpeechSynthesizer
from azure.cognitiveservices.speech.audio import AudioOutputConfig

from .azure_tts_voice import voice_name_to_language_code

lower2original = {k.lower(): k
                  for k in voice_name_to_language_code}


def determine_voice_name(voice_name):
    voice_name = voice_name.lower()
    if len(voice_name) == 0:
        name = 'en-US-BrandonNeural'
        language_code = 'en-US'
    else:
        candidates = list(filter(
            lambda lower_name: lower_name.startswith(voice_name),
            lower2original))
        if candidates:
            wavenet_candidates = list(
                filter(lambda c: 'wavenet' in c, candidates))
            voice_name = wavenet_candidates[0] \
                if wavenet_candidates else candidates[0]
            name = lower2original[voice_name]
            language_code = voice_name_to_language_code[name]
        else:
            raise RuntimeError(f'[Text2Wave] Invalid voice_name ({voice_name})')
    return name, language_code


def azure_text_to_speech(audio_filepath, text,
                         voice_name='en-US-BrandonNeural',
                         azure_key=None,
                         azure_region=None):
    audio_filepath = str(audio_filepath)
    if azure_key is None:
        azure_key = os.environ.get('AZURE_KEY', None)
        if azure_key is None:
            raise RuntimeError('AZURE_KEY is not specified.')
        azure_region = os.environ.get('AZURE_REGION', None)
        if azure_region is None:
            raise RuntimeError('AZURE_REGION is not specified.')
    voice_name, language_code = determine_voice_name(voice_name)

    speech_config = SpeechConfig(subscription=azure_key,
                                 region=azure_region)
    speech_config.speech_synthesis_voice_name = voice_name
    speech_config.request_word_level_timestamps()
    speech_config.speech_recognition_language = language_code

    audio_config = AudioOutputConfig(filename=audio_filepath)
    synthesizer = SpeechSynthesizer(speech_config=speech_config,
                                    audio_config=audio_config)
    synthesizer.speak_text_async(text)



voice_name_to_language_code = {
    'af-ZA-AdriNeural': 'af-ZA',  # Afrikaans (South Africa), Female
    'af-ZA-WillemNeural': 'af-ZA',  # Afrikaans (South Africa), Male
    'am-ET-MekdesNeural': 'am-ET',  # Amharic (Ethiopia), Female
    'am-ET-AmehaNeural': 'am-ET',  # Amharic (Ethiopia), Male
    'ar-DZ-AminaNeural': 'ar-DZ',  # Arabic (Algeria), Female
    'ar-DZ-IsmaelNeural': 'ar-DZ',  # Arabic (Algeria), Male
    'ar-BH-LailaNeural': 'ar-BH',  # Arabic (Bahrain), Female
    'ar-BH-AliNeural': 'ar-BH',  # Arabic (Bahrain), Male
    'ar-EG-SalmaNeural': 'ar-EG',  # Arabic (Egypt), Female
    'ar-EG-ShakirNeural': 'ar-EG',  # Arabic (Egypt), Male
    'ar-IQ-RanaNeural': 'ar-IQ',  # Arabic (Iraq), Female
    'ar-IQ-BasselNeural': 'ar-IQ',  # Arabic (Iraq), Male
    'ar-JO-SanaNeural': 'ar-JO',  # Arabic (Jordan), Female
    'ar-JO-TaimNeural': 'ar-JO',  # Arabic (Jordan), Male
    'ar-KW-NouraNeural': 'ar-KW',  # Arabic (Kuwait), Female
    'ar-KW-FahedNeural': 'ar-KW',  # Arabic (Kuwait), Male
    'ar-LY-ImanNeural': 'ar-LY',  # Arabic (Libya), Female
    'ar-LY-OmarNeural': 'ar-LY',  # Arabic (Libya), Male
    'ar-MA-MounaNeural': 'ar-MA',  # Arabic (Morocco), Female
    'ar-MA-JamalNeural': 'ar-MA',  # Arabic (Morocco), Male
    'ar-QA-AmalNeural': 'ar-QA',  # Arabic (Qatar), Female
    'ar-QA-MoazNeural': 'ar-QA',  # Arabic (Qatar), Male
    'ar-SA-ZariyahNeural': 'ar-SA',  # Arabic (Saudi Arabia), Female
    'ar-SA-HamedNeural': 'ar-SA',  # Arabic (Saudi Arabia), Male
    'ar-SY-AmanyNeural': 'ar-SY',  # Arabic (Syria), Female
    'ar-SY-LaithNeural': 'ar-SY',  # Arabic (Syria), Male
    'ar-TN-ReemNeural': 'ar-TN',  # Arabic (Tunisia), Female
    'ar-TN-HediNeural': 'ar-TN',  # Arabic (Tunisia), Male
    'ar-AE-FatimaNeural': 'ar-AE',  # Arabic (United Arab Emirates), Female
    'ar-AE-HamdanNeural': 'ar-AE',  # Arabic (United Arab Emirates), Male
    'ar-YE-MaryamNeural': 'ar-YE',  # Arabic (Yemen), Female
    'ar-YE-SalehNeural': 'ar-YE',  # Arabic (Yemen), Male
    'bn-BD-NabanitaNeural': 'bn-BD',  # Bangla (Bangladesh), Female
    'bn-BD-PradeepNeural': 'bn-BD',  # Bangla (Bangladesh), Male
    'bg-BG-KalinaNeural': 'bg-BG',  # Bulgarian (Bulgaria), Female
    'bg-BG-BorislavNeural': 'bg-BG',  # Bulgarian (Bulgaria), Male
    'my-MM-NilarNeural': 'my-MM',  # Burmese (Myanmar), Female
    'my-MM-ThihaNeural': 'my-MM',  # Burmese (Myanmar), Male
    'ca-ES-AlbaNeural': 'ca-ES',  # Catalan (Spain), Female
    'ca-ES-JoanaNeural': 'ca-ES',  # Catalan (Spain), Female
    'ca-ES-EnricNeural': 'ca-ES',  # Catalan (Spain), Male
    'zh-HK-HiuGaaiNeural': 'zh-HK',  # Chinese (Cantonese, Traditional), Female
    'zh-HK-HiuMaanNeural': 'zh-HK',  # Chinese (Cantonese, Traditional), Female
    'zh-HK-WanLungNeural': 'zh-HK',  # Chinese (Cantonese, Traditional), Male
    'zh-CN-XiaohanNeural': 'zh-CN',  # Chinese (Mandarin, Simplified), Female
    'zh-CN-XiaomoNeural': 'zh-CN',  # Chinese (Mandarin, Simplified), Female
    'zh-CN-XiaoruiNeural': 'zh-CN',  # Chinese (Mandarin, Simplified), Female
    'zh-CN-XiaoxiaoNeural': 'zh-CN',  # Chinese (Mandarin, Simplified), Female
    'zh-CN-XiaoxuanNeural': 'zh-CN',  # Chinese (Mandarin, Simplified), Female
    'zh-CN-XiaoyouNeural': 'zh-CN',  # Chinese (Mandarin, Simplified), Female
    'zh-CN-YunxiNeural': 'zh-CN',  # Chinese (Mandarin, Simplified), Male
    'zh-CN-YunyangNeural': 'zh-CN',  # Chinese (Mandarin, Simplified), Male
    'zh-CN-YunyeNeural': 'zh-CN',  # Chinese (Mandarin, Simplified), Male
    'zh-TW-HsiaoChenNeural': 'zh-TW',  # Chinese (Taiwanese Mandarin), Female
    'zh-TW-HsiaoYuNeural': 'zh-TW',  # Chinese (Taiwanese Mandarin), Female
    'zh-TW-YunJheNeural': 'zh-TW',  # Chinese (Taiwanese Mandarin), Male
    'hr-HR-GabrijelaNeural': 'hr-HR',  # Croatian (Croatia), Female
    'hr-HR-SreckoNeural': 'hr-HR',  # Croatian (Croatia), Male
    'cs-CZ-VlastaNeural': 'cs-CZ',  # Czech (Czech), Female
    'cs-CZ-AntoninNeural': 'cs-CZ',  # Czech (Czech), Male
    'da-DK-ChristelNeural': 'da-DK',  # Danish (Denmark), Female
    'da-DK-JeppeNeural': 'da-DK',  # Danish (Denmark), Male
    'nl-BE-DenaNeural': 'nl-BE',  # Dutch (Belgium), Female
    'nl-BE-ArnaudNeural': 'nl-BE',  # Dutch (Belgium), Male
    'nl-NL-ColetteNeural': 'nl-NL',  # Dutch (Netherlands), Female
    'nl-NL-FennaNeural': 'nl-NL',  # Dutch (Netherlands), Female
    'nl-NL-MaartenNeural': 'nl-NL',  # Dutch (Netherlands), Male
    'en-AU-NatashaNeural': 'en-AU',  # English (Australia), Female
    'en-AU-WilliamNeural': 'en-AU',  # English (Australia), Male
    'en-CA-ClaraNeural': 'en-CA',  # English (Canada), Female
    'en-CA-LiamNeural': 'en-CA',  # English (Canada), Male
    'en-HK-YanNeural': 'en-HK',  # English (Hongkong), Female
    'en-HK-SamNeural': 'en-HK',  # English (Hongkong), Male
    'en-IN-NeerjaNeural': 'en-IN',  # English (India), Female
    'en-IN-PrabhatNeural': 'en-IN',  # English (India), Male
    'en-IE-EmilyNeural': 'en-IE',  # English (Ireland), Female
    'en-IE-ConnorNeural': 'en-IE',  # English (Ireland), Male
    'en-KE-AsiliaNeural': 'en-KE',  # English (Kenya), Female
    'en-KE-ChilembaNeural': 'en-KE',  # English (Kenya), Male
    'en-NZ-MollyNeural': 'en-NZ',  # English (New Zealand), Female
    'en-NZ-MitchellNeural': 'en-NZ',  # English (New Zealand), Male
    'en-NG-EzinneNeural': 'en-NG',  # English (Nigeria), Female
    'en-NG-AbeoNeural': 'en-NG',  # English (Nigeria), Male
    'en-PH-RosaNeural': 'en-PH',  # English (Philippines), Female
    'en-PH-JamesNeural': 'en-PH',  # English (Philippines), Male
    'en-SG-LunaNeural': 'en-SG',  # English (Singapore), Female
    'en-SG-WayneNeural': 'en-SG',  # English (Singapore), Male
    'en-ZA-LeahNeural': 'en-ZA',  # English (South Africa), Female
    'en-ZA-LukeNeural': 'en-ZA',  # English (South Africa), Male
    'en-TZ-ImaniNeural': 'en-TZ',  # English (Tanzania), Female
    'en-TZ-ElimuNeural': 'en-TZ',  # English (Tanzania), Male
    'en-GB-LibbyNeural': 'en-GB',  # English (United Kingdom), Female
    'en-GB-SoniaNeural': 'en-GB',  # English (United Kingdom), Female
    'en-GB-MiaNeural': 'en-GB',  # English (United Kingdom), Female
    'en-GB-RyanNeural': 'en-GB',  # English (United Kingdom), Male
    'en-US-AmberNeural': 'en-US',  # English (United States), Female
    'en-US-AriaNeural': 'en-US',  # English (United States), Female
    'en-US-AshleyNeural': 'en-US',  # English (United States), Female
    'en-US-CoraNeural': 'en-US',  # English (United States), Female
    'en-US-ElizabethNeural': 'en-US',  # English (United States), Female
    'en-US-JennyNeural': 'en-US',  # English (United States), Female
    'en-US-MichelleNeural': 'en-US',  # English (United States), Female
    'en-US-MonicaNeural': 'en-US',  # English (United States), Female
    'en-US-SaraNeural': 'en-US',  # English (United States), Female
    'en-US-AnaNeural': 'en-US',  # English (United States), Kid
    'en-US-BrandonNeural': 'en-US',  # English (United States), Male
    'en-US-ChristopherNeural': 'en-US',  # English (United States), Male
    'en-US-EricNeural': 'en-US',  # English (United States), Male
    'en-US-GuyNeural': 'en-US',  # English (United States), Male
    'en-US-JacobNeural': 'en-US',  # English (United States), Male
    'et-EE-AnuNeural': 'et-EE',  # Estonian (Estonia), Female
    'et-EE-KertNeural': 'et-EE',  # Estonian (Estonia), Male
    'fil-PH-BlessicaNeural': 'fil-PH',  # Filipino (Philippines), Female
    'fil-PH-AngeloNeural': 'fil-PH',  # Filipino (Philippines), Male
    'fi-FI-NooraNeural': 'fi-FI',  # Finnish (Finland), Female
    'fi-FI-SelmaNeural': 'fi-FI',  # Finnish (Finland), Female
    'fi-FI-HarriNeural': 'fi-FI',  # Finnish (Finland), Male
    'fr-BE-CharlineNeural': 'fr-BE',  # French (Belgium), Female
    'fr-BE-GerardNeural': 'fr-BE',  # French (Belgium), Male
    'fr-CA-SylvieNeural': 'fr-CA',  # French (Canada), Female
    'fr-CA-AntoineNeural': 'fr-CA',  # French (Canada), Male
    'fr-CA-JeanNeural': 'fr-CA',  # French (Canada), Male
    'fr-FR-DeniseNeural': 'fr-FR',  # French (France), Female
    'fr-FR-HenriNeural': 'fr-FR',  # French (France), Male
    'fr-CH-ArianeNeural': 'fr-CH',  # French (Switzerland), Female
    'fr-CH-FabriceNeural': 'fr-CH',  # French (Switzerland), Male
    'gl-ES-SabelaNeural': 'gl-ES',  # Galician (Spain), Female
    'gl-ES-RoiNeural': 'gl-ES',  # Galician (Spain), Male
    'de-AT-IngridNeural': 'de-AT',  # German (Austria), Female
    'de-AT-JonasNeural': 'de-AT',  # German (Austria), Male
    'de-DE-KatjaNeural': 'de-DE',  # German (Germany), Female
    'de-DE-ConradNeural': 'de-DE',  # German (Germany), Male
    'de-CH-LeniNeural': 'de-CH',  # German (Switzerland), Female
    'de-CH-JanNeural': 'de-CH',  # German (Switzerland), Male
    'el-GR-AthinaNeural': 'el-GR',  # Greek (Greece), Female
    'el-GR-NestorasNeural': 'el-GR',  # Greek (Greece), Male
    'gu-IN-DhwaniNeural': 'gu-IN',  # Gujarati (India), Female
    'gu-IN-NiranjanNeural': 'gu-IN',  # Gujarati (India), Male
    'he-IL-HilaNeural': 'he-IL',  # Hebrew (Israel), Female
    'he-IL-AvriNeural': 'he-IL',  # Hebrew (Israel), Male
    'hi-IN-SwaraNeural': 'hi-IN',  # Hindi (India), Female
    'hi-IN-MadhurNeural': 'hi-IN',  # Hindi (India), Male
    'hu-HU-NoemiNeural': 'hu-HU',  # Hungarian (Hungary), Female
    'hu-HU-TamasNeural': 'hu-HU',  # Hungarian (Hungary), Male
    'id-ID-GadisNeural': 'id-ID',  # Indonesian (Indonesia), Female
    'id-ID-ArdiNeural': 'id-ID',  # Indonesian (Indonesia), Male
    'ga-IE-OrlaNeural': 'ga-IE',  # Irish (Ireland), Female
    'ga-IE-ColmNeural': 'ga-IE',  # Irish (Ireland), Male
    'it-IT-ElsaNeural': 'it-IT',  # Italian (Italy), Female
    'it-IT-IsabellaNeural': 'it-IT',  # Italian (Italy), Female
    'it-IT-DiegoNeural': 'it-IT',  # Italian (Italy), Male
    'ja-JP-NanamiNeural': 'ja-JP',  # Japanese (Japan), Female
    'ja-JP-KeitaNeural': 'ja-JP',  # Japanese (Japan), Male
    'jv-ID-SitiNeural': 'jv-ID',  # Javanese (Indonesia), Female
    'jv-ID-DimasNeural': 'jv-ID',  # Javanese (Indonesia), Male
    'km-KH-SreymomNeural': 'km-KH',  # Khmer (Cambodia), Female
    'km-KH-PisethNeural': 'km-KH',  # Khmer (Cambodia), Male
    'ko-KR-SunHiNeural': 'ko-KR',  # Korean (Korea), Female
    'ko-KR-InJoonNeural': 'ko-KR',  # Korean (Korea), Male
    'lv-LV-EveritaNeural': 'lv-LV',  # Latvian (Latvia), Female
    'lv-LV-NilsNeural': 'lv-LV',  # Latvian (Latvia), Male
    'lt-LT-OnaNeural': 'lt-LT',  # Lithuanian (Lithuania), Female
    'lt-LT-LeonasNeural': 'lt-LT',  # Lithuanian (Lithuania), Male
    'ms-MY-YasminNeural': 'ms-MY',  # Malay (Malaysia), Female
    'ms-MY-OsmanNeural': 'ms-MY',  # Malay (Malaysia), Male
    'mt-MT-GraceNeural': 'mt-MT',  # Maltese (Malta), Female
    'mt-MT-JosephNeural': 'mt-MT',  # Maltese (Malta), Male
    'mr-IN-AarohiNeural': 'mr-IN',  # Marathi (India), Female
    'mr-IN-ManoharNeural': 'mr-IN',  # Marathi (India), Male
    'nb-NO-IselinNeural': 'nb-NO',  # Norwegian (Bokml, Norway), Female
    'nb-NO-PernilleNeural': 'nb-NO',  # Norwegian (Bokml, Norway), Female
    'nb-NO-FinnNeural': 'nb-NO',  # Norwegian (Bokml, Norway), Male
    'fa-IR-DilaraNeural': 'fa-IR',  # Persian (Iran), Female
    'fa-IR-FaridNeural': 'fa-IR',  # Persian (Iran), Male
    'pl-PL-AgnieszkaNeural': 'pl-PL',  # Polish (Poland), Female
    'pl-PL-ZofiaNeural': 'pl-PL',  # Polish (Poland), Female
    'pl-PL-MarekNeural': 'pl-PL',  # Polish (Poland), Male
    'pt-BR-FranciscaNeural': 'pt-BR',  # Portuguese (Brazil), Female
    'pt-BR-AntonioNeural': 'pt-BR',  # Portuguese (Brazil), Male
    'pt-PT-FernandaNeural': 'pt-PT',  # Portuguese (Portugal), Female
    'pt-PT-RaquelNeural': 'pt-PT',  # Portuguese (Portugal), Female
    'pt-PT-DuarteNeural': 'pt-PT',  # Portuguese (Portugal), Male
    'ro-RO-AlinaNeural': 'ro-RO',  # Romanian (Romania), Female
    'ro-RO-EmilNeural': 'ro-RO',  # Romanian (Romania), Male
    'ru-RU-DariyaNeural': 'ru-RU',  # Russian (Russia), Female
    'ru-RU-SvetlanaNeural': 'ru-RU',  # Russian (Russia), Female
    'ru-RU-DmitryNeural': 'ru-RU',  # Russian (Russia), Male
    'sk-SK-ViktoriaNeural': 'sk-SK',  # Slovak (Slovakia), Female
    'sk-SK-LukasNeural': 'sk-SK',  # Slovak (Slovakia), Male
    'sl-SI-PetraNeural': 'sl-SI',  # Slovenian (Slovenia), Female
    'sl-SI-RokNeural': 'sl-SI',  # Slovenian (Slovenia), Male
    'so-SO-UbaxNeural': 'so-SO',  # Somali (Somalia), Female
    'so-SO-MuuseNeural': 'so-SO',  # Somali (Somalia), Male
    'es-AR-ElenaNeural': 'es-AR',  # Spanish (Argentina), Female
    'es-AR-TomasNeural': 'es-AR',  # Spanish (Argentina), Male
    'es-BO-SofiaNeural': 'es-BO',  # Spanish (Bolivia), Female
    'es-BO-MarceloNeural': 'es-BO',  # Spanish (Bolivia), Male
    'es-CL-CatalinaNeural': 'es-CL',  # Spanish (Chile), Female
    'es-CL-LorenzoNeural': 'es-CL',  # Spanish (Chile), Male
    'es-CO-SalomeNeural': 'es-CO',  # Spanish (Colombia), Female
    'es-CO-GonzaloNeural': 'es-CO',  # Spanish (Colombia), Male
    'es-CR-MariaNeural': 'es-CR',  # Spanish (Costa Rica), Female
    'es-CR-JuanNeural': 'es-CR',  # Spanish (Costa Rica), Male
    'es-CU-BelkysNeural': 'es-CU',  # Spanish (Cuba), Female
    'es-CU-ManuelNeural': 'es-CU',  # Spanish (Cuba), Male
    'es-DO-RamonaNeural': 'es-DO',  # Spanish (Dominican Republic), Female
    'es-DO-EmilioNeural': 'es-DO',  # Spanish (Dominican Republic), Male
    'es-EC-AndreaNeural': 'es-EC',  # Spanish (Ecuador), Female
    'es-EC-LuisNeural': 'es-EC',  # Spanish (Ecuador), Male
    'es-SV-LorenaNeural': 'es-SV',  # Spanish (El Salvador), Female
    'es-SV-RodrigoNeural': 'es-SV',  # Spanish (El Salvador), Male
    'es-GQ-TeresaNeural': 'es-GQ',  # Spanish (Equatorial Guinea), Female
    'es-GQ-JavierNeural': 'es-GQ',  # Spanish (Equatorial Guinea), Male
    'es-GT-MartaNeural': 'es-GT',  # Spanish (Guatemala), Female
    'es-GT-AndresNeural': 'es-GT',  # Spanish (Guatemala), Male
    'es-HN-KarlaNeural': 'es-HN',  # Spanish (Honduras), Female
    'es-HN-CarlosNeural': 'es-HN',  # Spanish (Honduras), Male
    'es-MX-DaliaNeural': 'es-MX',  # Spanish (Mexico), Female
    'es-MX-JorgeNeural': 'es-MX',  # Spanish (Mexico), Male
    'es-NI-YolandaNeural': 'es-NI',  # Spanish (Nicaragua), Female
    'es-NI-FedericoNeural': 'es-NI',  # Spanish (Nicaragua), Male
    'es-PA-MargaritaNeural': 'es-PA',  # Spanish (Panama), Female
    'es-PA-RobertoNeural': 'es-PA',  # Spanish (Panama), Male
    'es-PY-TaniaNeural': 'es-PY',  # Spanish (Paraguay), Female
    'es-PY-MarioNeural': 'es-PY',  # Spanish (Paraguay), Male
    'es-PE-CamilaNeural': 'es-PE',  # Spanish (Peru), Female
    'es-PE-AlexNeural': 'es-PE',  # Spanish (Peru), Male
    'es-PR-KarinaNeural': 'es-PR',  # Spanish (Puerto Rico), Female
    'es-PR-VictorNeural': 'es-PR',  # Spanish (Puerto Rico), Male
    'es-ES-ElviraNeural': 'es-ES',  # Spanish (Spain), Female
    'es-ES-AlvaroNeural': 'es-ES',  # Spanish (Spain), Male
    'es-UY-ValentinaNeural': 'es-UY',  # Spanish (Uruguay), Female
    'es-UY-MateoNeural': 'es-UY',  # Spanish (Uruguay), Male
    'es-US-PalomaNeural': 'es-US',  # Spanish (US), Female
    'es-US-AlonsoNeural': 'es-US',  # Spanish (US), Male
    'es-VE-PaolaNeural': 'es-VE',  # Spanish (Venezuela), Female
    'es-VE-SebastianNeural': 'es-VE',  # Spanish (Venezuela), Male
    'su-ID-TutiNeural': 'su-ID',  # Sundanese (Indonesia), Female
    'su-ID-JajangNeural': 'su-ID',  # Sundanese (Indonesia), Male
    'sw-KE-ZuriNeural': 'sw-KE',  # Swahili (Kenya), Female
    'sw-KE-RafikiNeural': 'sw-KE',  # Swahili (Kenya), Male
    'sw-TZ-RehemaNeural': 'sw-TZ',  # Swahili (Tanzania), Female
    'sw-TZ-DaudiNeural': 'sw-TZ',  # Swahili (Tanzania), Male
    'sv-SE-HilleviNeural': 'sv-SE',  # Swedish (Sweden), Female
    'sv-SE-SofieNeural': 'sv-SE',  # Swedish (Sweden), Female
    'sv-SE-MattiasNeural': 'sv-SE',  # Swedish (Sweden), Male
    'ta-IN-PallaviNeural': 'ta-IN',  # Tamil (India), Female
    'ta-IN-ValluvarNeural': 'ta-IN',  # Tamil (India), Male
    'ta-SG-VenbaNeural': 'ta-SG',  # Tamil (Singapore), Female
    'ta-SG-AnbuNeural': 'ta-SG',  # Tamil (Singapore), Male
    'ta-LK-SaranyaNeural': 'ta-LK',  # Tamil (Sri Lanka), Female
    'ta-LK-KumarNeural': 'ta-LK',  # Tamil (Sri Lanka), Male
    'te-IN-ShrutiNeural': 'te-IN',  # Telugu (India), Female
    'te-IN-MohanNeural': 'te-IN',  # Telugu (India), Male
    'th-TH-AcharaNeural': 'th-TH',  # Thai (Thailand), Female
    'th-TH-PremwadeeNeural': 'th-TH',  # Thai (Thailand), Female
    'th-TH-NiwatNeural': 'th-TH',  # Thai (Thailand), Male
    'tr-TR-EmelNeural': 'tr-TR',  # Turkish (Turkey), Female
    'tr-TR-AhmetNeural': 'tr-TR',  # Turkish (Turkey), Male
    'uk-UA-PolinaNeural': 'uk-UA',  # Ukrainian (Ukraine), Female
    'uk-UA-OstapNeural': 'uk-UA',  # Ukrainian (Ukraine), Male
    'ur-IN-GulNeural': 'ur-IN',  # Urdu (India), Female
    'ur-IN-SalmanNeural': 'ur-IN',  # Urdu (India), Male
    'ur-PK-UzmaNeural': 'ur-PK',  # Urdu (Pakistan), Female
    'ur-PK-AsadNeural': 'ur-PK',  # Urdu (Pakistan), Male
    'uz-UZ-MadinaNeural': 'uz-UZ',  # Uzbek (Uzbekistan), Female
    'uz-UZ-SardorNeural': 'uz-UZ',  # Uzbek (Uzbekistan), Male
    'vi-VN-HoaiMyNeural': 'vi-VN',  # Vietnamese (Vietnam), Female
    'vi-VN-NamMinhNeural': 'vi-VN',  # Vietnamese (Vietnam), Male
    'cy-GB-NiaNeural': 'cy-GB',  # Welsh (United Kingdom), Female
    'cy-GB-AledNeural': 'cy-GB',  # Welsh (United Kingdom), Male
    'zu-ZA-ThandoNeural': 'zu-ZA',  # Zulu (South Africa), Female
    'zu-ZA-ThembaNeural': 'zu-ZA',  # Zulu (South Africa), Male
}



from pathlib import Path

from google.cloud import texttospeech
import pydub

from .google_tts_voice import voice_name_to_language_code

lower2original = {k.lower(): k
                  for k in voice_name_to_language_code}


def determine_voice_name(voice_name):
    voice_name = voice_name.lower()
    if len(voice_name) == 0:
        name = 'en-US-Wavenet-A'
        language_code = 'en-US'
    else:
        candidates = list(filter(
            lambda lower_name: lower_name.startswith(voice_name),
            lower2original))
        if candidates:
            wavenet_candidates = list(
                filter(lambda c: 'wavenet' in c, candidates))
            voice_name = wavenet_candidates[0] \
                if wavenet_candidates else candidates[0]
            name = lower2original[voice_name]
            language_code = voice_name_to_language_code[name]
        else:
            raise RuntimeError(f'[Text2Wave] Invalid voice_name ({voice_name})')
    return name, language_code


def google_text_to_speech(
        audio_filepath, text,
        voice_name='en-US-Wavenet-F',
        speaking_rate=1.0,
        sample_rate=16000):
    """Text to speech.

    You can see which language is available here
    https://cloud.google.com/text-to-speech/docs/voices
    """
    audio_filepath = Path(audio_filepath)
    voice_name, language_code = determine_voice_name(voice_name)

    client = texttospeech.TextToSpeechClient()
    synthesis_input = texttospeech.SynthesisInput(
        text=text)
    voice = texttospeech.VoiceSelectionParams(
        language_code=language_code,
        name=voice_name,
        ssml_gender=texttospeech.SsmlVoiceGender.FEMALE)

    # Select the type of audio file you want returned
    audio_config = texttospeech.AudioConfig(
        audio_encoding=texttospeech.AudioEncoding.MP3,
        sample_rate_hertz=sample_rate,
        speaking_rate=speaking_rate)
    response = client.synthesize_speech(
        input=synthesis_input,
        voice=voice,
        audio_config=audio_config)

    with open(str(audio_filepath.with_suffix('.mp3')), 'wb') as out:
        out.write(response.audio_content)

    if audio_filepath.suffix == '.wav':
        sound = pydub.AudioSegment.from_mp3(
            str(audio_filepath.with_suffix('.mp3')))
        sound.export(str(audio_filepath), format="wav")



voice_name_to_language_code = {
    'af-ZA-Standard-A': 'af-ZA',  # Afrikaans (South Africa), FEMALE
    'ar-XA-Standard-A': 'ar-XA',  # Arabic, FEMALE
    'ar-XA-Standard-B': 'ar-XA',  # Arabic, MALE
    'ar-XA-Standard-C': 'ar-XA',  # Arabic, MALE
    'ar-XA-Standard-D': 'ar-XA',  # Arabic, FEMALE
    'ar-XA-Wavenet-A': 'ar-XA',  # Arabic, FEMALE
    'ar-XA-Wavenet-B': 'ar-XA',  # Arabic, MALE
    'ar-XA-Wavenet-C': 'ar-XA',  # Arabic, MALE
    'ar-XA-Wavenet-D': 'ar-XA',  # Arabic, FEMALE
    'bn-IN-Standard-A': 'bn-IN',  # Bengali (India), FEMALE
    'bn-IN-Standard-B': 'bn-IN',  # Bengali (India), MALE
    'bn-IN-Wavenet-A': 'bn-IN',  # Bengali (India), FEMALE
    'bn-IN-Wavenet-B': 'bn-IN',  # Bengali (India), MALE
    'bg-bg-Standard-A': 'bg-BG',  # Bulgarian (Bulgaria), FEMALE
    'ca-es-Standard-A': 'ca-ES',  # Catalan (Spain), FEMALE
    'yue-HK-Standard-A': 'yue-HK',  # Chinese (Hong Kong), FEMALE
    'yue-HK-Standard-B': 'yue-HK',  # Chinese (Hong Kong), MALE
    'yue-HK-Standard-C': 'yue-HK',  # Chinese (Hong Kong), FEMALE
    'yue-HK-Standard-D': 'yue-HK',  # Chinese (Hong Kong), MALE
    'cs-CZ-Standard-A': 'cs-CZ',  # Czech (Czech Republic), FEMALE
    'cs-CZ-Wavenet-A': 'cs-CZ',  # Czech (Czech Republic), FEMALE
    'da-DK-Standard-A': 'da-DK',  # Danish (Denmark), FEMALE
    'da-DK-Standard-C': 'da-DK',  # Danish (Denmark), MALE
    'da-DK-Standard-D': 'da-DK',  # Danish (Denmark), FEMALE
    'da-DK-Standard-E': 'da-DK',  # Danish (Denmark), FEMALE
    'da-DK-Wavenet-A': 'da-DK',  # Danish (Denmark), FEMALE
    'da-DK-Wavenet-C': 'da-DK',  # Danish (Denmark), MALE
    'da-DK-Wavenet-D': 'da-DK',  # Danish (Denmark), FEMALE
    'da-DK-Wavenet-E': 'da-DK',  # Danish (Denmark), FEMALE
    'nl-BE-Standard-A': 'nl-BE',  # Dutch (Belgium), FEMALE
    'nl-BE-Standard-B': 'nl-BE',  # Dutch (Belgium), MALE
    'nl-BE-Wavenet-A': 'nl-BE',  # Dutch (Belgium), FEMALE
    'nl-BE-Wavenet-B': 'nl-BE',  # Dutch (Belgium), MALE
    'nl-NL-Standard-A': 'nl-NL',  # Dutch (Netherlands), FEMALE
    'nl-NL-Standard-B': 'nl-NL',  # Dutch (Netherlands), MALE
    'nl-NL-Standard-C': 'nl-NL',  # Dutch (Netherlands), MALE
    'nl-NL-Standard-D': 'nl-NL',  # Dutch (Netherlands), FEMALE
    'nl-NL-Standard-E': 'nl-NL',  # Dutch (Netherlands), FEMALE
    'nl-NL-Wavenet-A': 'nl-NL',  # Dutch (Netherlands), FEMALE
    'nl-NL-Wavenet-B': 'nl-NL',  # Dutch (Netherlands), MALE
    'nl-NL-Wavenet-C': 'nl-NL',  # Dutch (Netherlands), MALE
    'nl-NL-Wavenet-D': 'nl-NL',  # Dutch (Netherlands), FEMALE
    'nl-NL-Wavenet-E': 'nl-NL',  # Dutch (Netherlands), FEMALE
    'en-AU-Standard-A': 'en-AU',  # English (Australia), FEMALE
    'en-AU-Standard-B': 'en-AU',  # English (Australia), MALE
    'en-AU-Standard-C': 'en-AU',  # English (Australia), FEMALE
    'en-AU-Standard-D': 'en-AU',  # English (Australia), MALE
    'en-AU-Wavenet-A': 'en-AU',  # English (Australia), FEMALE
    'en-AU-Wavenet-B': 'en-AU',  # English (Australia), MALE
    'en-AU-Wavenet-C': 'en-AU',  # English (Australia), FEMALE
    'en-AU-Wavenet-D': 'en-AU',  # English (Australia), MALE
    'en-IN-Standard-A': 'en-IN',  # English (India), FEMALE
    'en-IN-Standard-B': 'en-IN',  # English (India), MALE
    'en-IN-Standard-C': 'en-IN',  # English (India), MALE
    'en-IN-Standard-D': 'en-IN',  # English (India), FEMALE
    'en-IN-Wavenet-A': 'en-IN',  # English (India), FEMALE
    'en-IN-Wavenet-B': 'en-IN',  # English (India), MALE
    'en-IN-Wavenet-C': 'en-IN',  # English (India), MALE
    'en-IN-Wavenet-D': 'en-IN',  # English (India), FEMALE
    'en-GB-Standard-A': 'en-GB',  # English (UK), FEMALE
    'en-GB-Standard-B': 'en-GB',  # English (UK), MALE
    'en-GB-Standard-C': 'en-GB',  # English (UK), FEMALE
    'en-GB-Standard-D': 'en-GB',  # English (UK), MALE
    'en-GB-Standard-F': 'en-GB',  # English (UK), FEMALE
    'en-GB-Wavenet-A': 'en-GB',  # English (UK), FEMALE
    'en-GB-Wavenet-B': 'en-GB',  # English (UK), MALE
    'en-GB-Wavenet-C': 'en-GB',  # English (UK), FEMALE
    'en-GB-Wavenet-D': 'en-GB',  # English (UK), MALE
    'en-GB-Wavenet-F': 'en-GB',  # English (UK), FEMALE
    'en-US-Standard-A': 'en-US',  # English (US), MALE
    'en-US-Standard-B': 'en-US',  # English (US), MALE
    'en-US-Standard-C': 'en-US',  # English (US), FEMALE
    'en-US-Standard-D': 'en-US',  # English (US), MALE
    'en-US-Standard-E': 'en-US',  # English (US), FEMALE
    'en-US-Standard-F': 'en-US',  # English (US), FEMALE
    'en-US-Standard-G': 'en-US',  # English (US), FEMALE
    'en-US-Standard-H': 'en-US',  # English (US), FEMALE
    'en-US-Standard-I': 'en-US',  # English (US), MALE
    'en-US-Standard-J': 'en-US',  # English (US), MALE
    'en-US-Wavenet-A': 'en-US',  # English (US), MALE
    'en-US-Wavenet-B': 'en-US',  # English (US), MALE
    'en-US-Wavenet-C': 'en-US',  # English (US), FEMALE
    'en-US-Wavenet-D': 'en-US',  # English (US), MALE
    'en-US-Wavenet-E': 'en-US',  # English (US), FEMALE
    'en-US-Wavenet-F': 'en-US',  # English (US), FEMALE
    'en-US-Wavenet-G': 'en-US',  # English (US), FEMALE
    'en-US-Wavenet-H': 'en-US',  # English (US), FEMALE
    'en-US-Wavenet-I': 'en-US',  # English (US), MALE
    'en-US-Wavenet-J': 'en-US',  # English (US), MALE
    'fil-PH-Standard-A': 'fil-PH',  # Filipino (Philippines), FEMALE
    'fil-PH-Standard-B': 'fil-PH',  # Filipino (Philippines), FEMALE
    'fil-PH-Standard-C': 'fil-PH',  # Filipino (Philippines), MALE
    'fil-PH-Standard-D': 'fil-PH',  # Filipino (Philippines), MALE
    'fil-PH-Wavenet-A': 'fil-PH',  # Filipino (Philippines), FEMALE
    'fil-PH-Wavenet-B': 'fil-PH',  # Filipino (Philippines), FEMALE
    'fil-PH-Wavenet-C': 'fil-PH',  # Filipino (Philippines), MALE
    'fil-PH-Wavenet-D': 'fil-PH',  # Filipino (Philippines), MALE
    'fi-FI-Standard-A': 'fi-FI',  # Finnish (Finland), FEMALE
    'fi-FI-Wavenet-A': 'fi-FI',  # Finnish (Finland), FEMALE
    'fr-CA-Standard-A': 'fr-CA',  # French (Canada), FEMALE
    'fr-CA-Standard-B': 'fr-CA',  # French (Canada), MALE
    'fr-CA-Standard-C': 'fr-CA',  # French (Canada), FEMALE
    'fr-CA-Standard-D': 'fr-CA',  # French (Canada), MALE
    'fr-CA-Wavenet-A': 'fr-CA',  # French (Canada), FEMALE
    'fr-CA-Wavenet-B': 'fr-CA',  # French (Canada), MALE
    'fr-CA-Wavenet-C': 'fr-CA',  # French (Canada), FEMALE
    'fr-CA-Wavenet-D': 'fr-CA',  # French (Canada), MALE
    'fr-FR-Standard-A': 'fr-FR',  # French (France), FEMALE
    'fr-FR-Standard-B': 'fr-FR',  # French (France), MALE
    'fr-FR-Standard-C': 'fr-FR',  # French (France), FEMALE
    'fr-FR-Standard-D': 'fr-FR',  # French (France), MALE
    'fr-FR-Standard-E': 'fr-FR',  # French (France), FEMALE
    'fr-FR-Wavenet-A': 'fr-FR',  # French (France), FEMALE
    'fr-FR-Wavenet-B': 'fr-FR',  # French (France), MALE
    'fr-FR-Wavenet-C': 'fr-FR',  # French (France), FEMALE
    'fr-FR-Wavenet-D': 'fr-FR',  # French (France), MALE
    'fr-FR-Wavenet-E': 'fr-FR',  # French (France), FEMALE
    'de-DE-Standard-A': 'de-DE',  # German (Germany), FEMALE
    'de-DE-Standard-B': 'de-DE',  # German (Germany), MALE
    'de-DE-Standard-C': 'de-DE',  # German (Germany), FEMALE
    'de-DE-Standard-D': 'de-DE',  # German (Germany), MALE
    'de-DE-Standard-E': 'de-DE',  # German (Germany), MALE
    'de-DE-Standard-F': 'de-DE',  # German (Germany), FEMALE
    'de-DE-Wavenet-A': 'de-DE',  # German (Germany), FEMALE
    'de-DE-Wavenet-B': 'de-DE',  # German (Germany), MALE
    'de-DE-Wavenet-C': 'de-DE',  # German (Germany), FEMALE
    'de-DE-Wavenet-D': 'de-DE',  # German (Germany), MALE
    'de-DE-Wavenet-E': 'de-DE',  # German (Germany), MALE
    'de-DE-Wavenet-F': 'de-DE',  # German (Germany), FEMALE
    'el-GR-Standard-A': 'el-GR',  # Greek (Greece), FEMALE
    'el-GR-Wavenet-A': 'el-GR',  # Greek (Greece), FEMALE
    'gu-IN-Standard-A': 'gu-IN',  # Gujarati (India), FEMALE
    'gu-IN-Standard-B': 'gu-IN',  # Gujarati (India), MALE
    'gu-IN-Wavenet-A': 'gu-IN',  # Gujarati (India), FEMALE
    'gu-IN-Wavenet-B': 'gu-IN',  # Gujarati (India), MALE
    'hi-IN-Standard-A': 'hi-IN',  # Hindi (India), FEMALE
    'hi-IN-Standard-B': 'hi-IN',  # Hindi (India), MALE
    'hi-IN-Standard-C': 'hi-IN',  # Hindi (India), MALE
    'hi-IN-Standard-D': 'hi-IN',  # Hindi (India), FEMALE
    'hi-IN-Wavenet-A': 'hi-IN',  # Hindi (India), FEMALE
    'hi-IN-Wavenet-B': 'hi-IN',  # Hindi (India), MALE
    'hi-IN-Wavenet-C': 'hi-IN',  # Hindi (India), MALE
    'hi-IN-Wavenet-D': 'hi-IN',  # Hindi (India), FEMALE
    'hu-HU-Standard-A': 'hu-HU',  # Hungarian (Hungary), FEMALE
    'hu-HU-Wavenet-A': 'hu-HU',  # Hungarian (Hungary), FEMALE
    'is-is-Standard-A': 'is-IS',  # Icelandic (Iceland), FEMALE
    'id-ID-Standard-A': 'id-ID',  # Indonesian (Indonesia), FEMALE
    'id-ID-Standard-B': 'id-ID',  # Indonesian (Indonesia), MALE
    'id-ID-Standard-C': 'id-ID',  # Indonesian (Indonesia), MALE
    'id-ID-Standard-D': 'id-ID',  # Indonesian (Indonesia), FEMALE
    'id-ID-Wavenet-A': 'id-ID',  # Indonesian (Indonesia), FEMALE
    'id-ID-Wavenet-B': 'id-ID',  # Indonesian (Indonesia), MALE
    'id-ID-Wavenet-C': 'id-ID',  # Indonesian (Indonesia), MALE
    'id-ID-Wavenet-D': 'id-ID',  # Indonesian (Indonesia), FEMALE
    'it-IT-Standard-A': 'it-IT',  # Italian (Italy), FEMALE
    'it-IT-Standard-B': 'it-IT',  # Italian (Italy), FEMALE
    'it-IT-Standard-C': 'it-IT',  # Italian (Italy), MALE
    'it-IT-Standard-D': 'it-IT',  # Italian (Italy), MALE
    'it-IT-Wavenet-A': 'it-IT',  # Italian (Italy), FEMALE
    'it-IT-Wavenet-B': 'it-IT',  # Italian (Italy), FEMALE
    'it-IT-Wavenet-C': 'it-IT',  # Italian (Italy), MALE
    'it-IT-Wavenet-D': 'it-IT',  # Italian (Italy), MALE
    'ja-JP-Standard-A': 'ja-JP',  # Japanese (Japan), FEMALE
    'ja-JP-Standard-B': 'ja-JP',  # Japanese (Japan), FEMALE
    'ja-JP-Standard-C': 'ja-JP',  # Japanese (Japan), MALE
    'ja-JP-Standard-D': 'ja-JP',  # Japanese (Japan), MALE
    'ja-JP-Wavenet-A': 'ja-JP',  # Japanese (Japan), FEMALE
    'ja-JP-Wavenet-B': 'ja-JP',  # Japanese (Japan), FEMALE
    'ja-JP-Wavenet-C': 'ja-JP',  # Japanese (Japan), MALE
    'ja-JP-Wavenet-D': 'ja-JP',  # Japanese (Japan), MALE
    'kn-IN-Standard-A': 'kn-IN',  # Kannada (India), FEMALE
    'kn-IN-Standard-B': 'kn-IN',  # Kannada (India), MALE
    'kn-IN-Wavenet-A': 'kn-IN',  # Kannada (India), FEMALE
    'kn-IN-Wavenet-B': 'kn-IN',  # Kannada (India), MALE
    'ko-KR-Standard-A': 'ko-KR',  # Korean (South Korea), FEMALE
    'ko-KR-Standard-B': 'ko-KR',  # Korean (South Korea), FEMALE
    'ko-KR-Standard-C': 'ko-KR',  # Korean (South Korea), MALE
    'ko-KR-Standard-D': 'ko-KR',  # Korean (South Korea), MALE
    'ko-KR-Wavenet-A': 'ko-KR',  # Korean (South Korea), FEMALE
    'ko-KR-Wavenet-B': 'ko-KR',  # Korean (South Korea), FEMALE
    'ko-KR-Wavenet-C': 'ko-KR',  # Korean (South Korea), MALE
    'ko-KR-Wavenet-D': 'ko-KR',  # Korean (South Korea), MALE
    'lv-lv-Standard-A': 'lv-LV',  # Latvian (Latvia), MALE
    'ms-MY-Standard-A': 'ms-MY',  # Malay (Malaysia), FEMALE
    'ms-MY-Standard-B': 'ms-MY',  # Malay (Malaysia), MALE
    'ms-MY-Standard-C': 'ms-MY',  # Malay (Malaysia), FEMALE
    'ms-MY-Standard-D': 'ms-MY',  # Malay (Malaysia), MALE
    'ms-MY-Wavenet-A': 'ms-MY',  # Malay (Malaysia), FEMALE
    'ms-MY-Wavenet-B': 'ms-MY',  # Malay (Malaysia), MALE
    'ms-MY-Wavenet-C': 'ms-MY',  # Malay (Malaysia), FEMALE
    'ms-MY-Wavenet-D': 'ms-MY',  # Malay (Malaysia), MALE
    'ml-IN-Standard-A': 'ml-IN',  # Malayalam (India), FEMALE
    'ml-IN-Standard-B': 'ml-IN',  # Malayalam (India), MALE
    'ml-IN-Wavenet-A': 'ml-IN',  # Malayalam (India), FEMALE
    'ml-IN-Wavenet-B': 'ml-IN',  # Malayalam (India), MALE
    'cmn-CN-Standard-A': 'cmn-CN',  # Mandarin Chinese, FEMALE
    'cmn-CN-Standard-B': 'cmn-CN',  # Mandarin Chinese, MALE
    'cmn-CN-Standard-C': 'cmn-CN',  # Mandarin Chinese, MALE
    'cmn-CN-Standard-D': 'cmn-CN',  # Mandarin Chinese, FEMALE
    'cmn-CN-Wavenet-A': 'cmn-CN',  # Mandarin Chinese, FEMALE
    'cmn-CN-Wavenet-B': 'cmn-CN',  # Mandarin Chinese, MALE
    'cmn-CN-Wavenet-C': 'cmn-CN',  # Mandarin Chinese, MALE
    'cmn-CN-Wavenet-D': 'cmn-CN',  # Mandarin Chinese, FEMALE
    'cmn-TW-Standard-A': 'cmn-TW',  # Mandarin Chinese, FEMALE
    'cmn-TW-Standard-B': 'cmn-TW',  # Mandarin Chinese, MALE
    'cmn-TW-Standard-C': 'cmn-TW',  # Mandarin Chinese, MALE
    'cmn-TW-Wavenet-A': 'cmn-TW',  # Mandarin Chinese, FEMALE
    'cmn-TW-Wavenet-B': 'cmn-TW',  # Mandarin Chinese, MALE
    'cmn-TW-Wavenet-C': 'cmn-TW',  # Mandarin Chinese, MALE
    'nb-NO-Standard-A': 'nb-NO',  # Norwegian (Norway), FEMALE
    'nb-NO-Standard-B': 'nb-NO',  # Norwegian (Norway), MALE
    'nb-NO-Standard-C': 'nb-NO',  # Norwegian (Norway), FEMALE
    'nb-NO-Standard-D': 'nb-NO',  # Norwegian (Norway), MALE
    'nb-NO-Wavenet-A': 'nb-NO',  # Norwegian (Norway), FEMALE
    'nb-NO-Wavenet-B': 'nb-NO',  # Norwegian (Norway), MALE
    'nb-NO-Wavenet-C': 'nb-NO',  # Norwegian (Norway), FEMALE
    'nb-NO-Wavenet-D': 'nb-NO',  # Norwegian (Norway), MALE
    'nb-no-Standard-E': 'nb-NO',  # Norwegian (Norway), FEMALE
    'nb-no-Wavenet-E': 'nb-NO',  # Norwegian (Norway), FEMALE
    'pl-PL-Standard-A': 'pl-PL',  # Polish (Poland), FEMALE
    'pl-PL-Standard-B': 'pl-PL',  # Polish (Poland), MALE
    'pl-PL-Standard-C': 'pl-PL',  # Polish (Poland), MALE
    'pl-PL-Standard-D': 'pl-PL',  # Polish (Poland), FEMALE
    'pl-PL-Standard-E': 'pl-PL',  # Polish (Poland), FEMALE
    'pl-PL-Wavenet-A': 'pl-PL',  # Polish (Poland), FEMALE
    'pl-PL-Wavenet-B': 'pl-PL',  # Polish (Poland), MALE
    'pl-PL-Wavenet-C': 'pl-PL',  # Polish (Poland), MALE
    'pl-PL-Wavenet-D': 'pl-PL',  # Polish (Poland), FEMALE
    'pl-PL-Wavenet-E': 'pl-PL',  # Polish (Poland), FEMALE
    'pt-BR-Standard-A': 'pt-BR',  # Portuguese (Brazil), FEMALE
    'pt-BR-Standard-B': 'pt-BR',  # Portuguese (Brazil), MALE
    'pt-BR-Wavenet-A': 'pt-BR',  # Portuguese (Brazil), FEMALE
    'pt-BR-Wavenet-B': 'pt-BR',  # Portuguese (Brazil), MALE
    'pt-PT-Standard-A': 'pt-PT',  # Portuguese (Portugal), FEMALE
    'pt-PT-Standard-B': 'pt-PT',  # Portuguese (Portugal), MALE
    'pt-PT-Standard-C': 'pt-PT',  # Portuguese (Portugal), MALE
    'pt-PT-Standard-D': 'pt-PT',  # Portuguese (Portugal), FEMALE
    'pt-PT-Wavenet-A': 'pt-PT',  # Portuguese (Portugal), FEMALE
    'pt-PT-Wavenet-B': 'pt-PT',  # Portuguese (Portugal), MALE
    'pt-PT-Wavenet-C': 'pt-PT',  # Portuguese (Portugal), MALE
    'pt-PT-Wavenet-D': 'pt-PT',  # Portuguese (Portugal), FEMALE
    'pa-IN-Standard-A': 'pa-IN',  # Punjabi (India), FEMALE
    'pa-IN-Standard-B': 'pa-IN',  # Punjabi (India), MALE
    'pa-IN-Standard-C': 'pa-IN',  # Punjabi (India), FEMALE
    'pa-IN-Standard-D': 'pa-IN',  # Punjabi (India), MALE
    'pa-IN-Wavenet-A': 'pa-IN',  # Punjabi (India), FEMALE
    'pa-IN-Wavenet-B': 'pa-IN',  # Punjabi (India), MALE
    'pa-IN-Wavenet-C': 'pa-IN',  # Punjabi (India), FEMALE
    'pa-IN-Wavenet-D': 'pa-IN',  # Punjabi (India), MALE
    'ro-RO-Standard-A': 'ro-RO',  # Romanian (Romania), FEMALE
    'ro-RO-Wavenet-A': 'ro-RO',  # Romanian (Romania), FEMALE
    'ru-RU-Standard-A': 'ru-RU',  # Russian (Russia), FEMALE
    'ru-RU-Standard-B': 'ru-RU',  # Russian (Russia), MALE
    'ru-RU-Standard-C': 'ru-RU',  # Russian (Russia), FEMALE
    'ru-RU-Standard-D': 'ru-RU',  # Russian (Russia), MALE
    'ru-RU-Standard-E': 'ru-RU',  # Russian (Russia), FEMALE
    'ru-RU-Wavenet-A': 'ru-RU',  # Russian (Russia), FEMALE
    'ru-RU-Wavenet-B': 'ru-RU',  # Russian (Russia), MALE
    'ru-RU-Wavenet-C': 'ru-RU',  # Russian (Russia), FEMALE
    'ru-RU-Wavenet-D': 'ru-RU',  # Russian (Russia), MALE
    'ru-RU-Wavenet-E': 'ru-RU',  # Russian (Russia), FEMALE
    'sr-rs-Standard-A': 'sr-RS',  # Serbian (Cyrillic), FEMALE
    'sk-SK-Standard-A': 'sk-SK',  # Slovak (Slovakia), FEMALE
    'sk-SK-Wavenet-A': 'sk-SK',  # Slovak (Slovakia), FEMALE
    'es-ES-Standard-A': 'es-ES',  # Spanish (Spain), FEMALE
    'es-ES-Standard-B': 'es-ES',  # Spanish (Spain), MALE
    'es-ES-Standard-C': 'es-ES',  # Spanish (Spain), FEMALE
    'es-ES-Standard-D': 'es-ES',  # Spanish (Spain), FEMALE
    'es-ES-Wavenet-B': 'es-ES',  # Spanish (Spain), MALE
    'es-ES-Wavenet-C': 'es-ES',  # Spanish (Spain), FEMALE
    'es-ES-Wavenet-D': 'es-ES',  # Spanish (Spain), FEMALE
    'es-US-Standard-A': 'es-US',  # Spanish (US), FEMALE
    'es-US-Standard-B': 'es-US',  # Spanish (US), MALE
    'es-US-Standard-C': 'es-US',  # Spanish (US), MALE
    'es-US-Wavenet-A': 'es-US',  # Spanish (US), FEMALE
    'es-US-Wavenet-B': 'es-US',  # Spanish (US), MALE
    'es-US-Wavenet-C': 'es-US',  # Spanish (US), MALE
    'sv-SE-Standard-A': 'sv-SE',  # Swedish (Sweden), FEMALE
    'sv-SE-Wavenet-A': 'sv-SE',  # Swedish (Sweden), FEMALE
    'ta-IN-Standard-A': 'ta-IN',  # Tamil (India), FEMALE
    'ta-IN-Standard-B': 'ta-IN',  # Tamil (India), MALE
    'ta-IN-Wavenet-A': 'ta-IN',  # Tamil (India), FEMALE
    'ta-IN-Wavenet-B': 'ta-IN',  # Tamil (India), MALE
    'te-IN-Standard-A': 'te-IN',  # Telugu (India), FEMALE
    'te-IN-Standard-B': 'te-IN',  # Telugu (India), MALE
    'th-TH-Standard-A': 'th-TH',  # Thai (Thailand), FEMALE
    'tr-TR-Standard-A': 'tr-TR',  # Turkish (Turkey), FEMALE
    'tr-TR-Standard-B': 'tr-TR',  # Turkish (Turkey), MALE
    'tr-TR-Standard-C': 'tr-TR',  # Turkish (Turkey), FEMALE
    'tr-TR-Standard-D': 'tr-TR',  # Turkish (Turkey), FEMALE
    'tr-TR-Standard-E': 'tr-TR',  # Turkish (Turkey), MALE
    'tr-TR-Wavenet-A': 'tr-TR',  # Turkish (Turkey), FEMALE
    'tr-TR-Wavenet-B': 'tr-TR',  # Turkish (Turkey), MALE
    'tr-TR-Wavenet-C': 'tr-TR',  # Turkish (Turkey), FEMALE
    'tr-TR-Wavenet-D': 'tr-TR',  # Turkish (Turkey), FEMALE
    'tr-TR-Wavenet-E': 'tr-TR',  # Turkish (Turkey), MALE
    'uk-UA-Standard-A': 'uk-UA',  # Ukrainian (Ukraine), FEMALE
    'uk-UA-Wavenet-A': 'uk-UA',  # Ukrainian (Ukraine), FEMALE
    'vi-VN-Standard-A': 'vi-VN',  # Vietnamese (Vietnam), FEMALE
    'vi-VN-Standard-B': 'vi-VN',  # Vietnamese (Vietnam), MALE
    'vi-VN-Standard-C': 'vi-VN',  # Vietnamese (Vietnam), FEMALE
    'vi-VN-Standard-D': 'vi-VN',  # Vietnamese (Vietnam), MALE
    'vi-VN-Wavenet-A': 'vi-VN',  # Vietnamese (Vietnam), FEMALE
    'vi-VN-Wavenet-B': 'vi-VN',  # Vietnamese (Vietnam), MALE
    'vi-VN-Wavenet-C': 'vi-VN',  # Vietnamese (Vietnam), FEMALE
    'vi-VN-Wavenet-D': 'vi-VN',  # Vietnamese (Vietnam), MALE
}



import re

valid = True
try:
    import urllib
    from urllib.request import urlopen
except ImportError:
    valid = False


def download_file(url, dst_path):
    """Download file from url.

    Parameters
    ----------
    url : str
        url.
    dst_path : str
        destination path.
    """
    if valid is False:
        raise RuntimeError('Not supported.')
    dst_path = str(dst_path)
    regex = r'[^\x00-\x7F]'
    matchedList = re.findall(regex, url)
    for m in matchedList:
        url = url.replace(m, urllib.parse.quote_plus(m, encoding="utf-8"))
    with urlopen(url) as web_file:
        data = web_file.read()
        with open(dst_path, mode='wb') as local_file:
            local_file.write(data)




import random

import numpy as np


def coloring(img, w_i, h_i, check_w, check_h, color_n, color1, color2):
    if color_n % 2 == 0:
        c = color1
    else:
        c = color2
    max_w, max_h = img.shape[:2]
    w_s = check_w * w_i
    w_e = check_w * (w_i + 1)
    if w_e > max_w:
        w_e = max_w
    h_s = check_h * h_i
    h_e = check_h * (h_i + 1)
    if h_e > max_h:
        h_e = max_h

    img[w_s:w_e, h_s:h_e, :] = c
    return img


def generate_single_color_texture(size_wh=(300, 300)):
    w, h = list(map(int, size_wh))
    color = [np.random.randint(0, 255), np.random.randint(
        0, 255), np.random.randint(0, 255)]
    np_img = np.array([color for _ in range(w*h)]).astype(np.uint8)
    np_img = np.reshape(np_img, (w, h, 3))
    return np_img


def generate_check_texture(size_wh=(300, 300)):
    w, h = list(map(int, size_wh))
    np_img = np.zeros([w, h, 3]).astype(np.uint8)
    color_n = 0
    color1 = [np.random.randint(0, 255), np.random.randint(
        0, 255), np.random.randint(0, 255)]
    color2 = [np.random.randint(0, 255), np.random.randint(
        0, 255), np.random.randint(0, 255)]
    check_w = np.random.randint(1, w / 2)
    check_h = np.random.randint(1, h / 2)
    w_iter = int(w / check_w)
    h_iter = int(h / check_h)
    for wi in range(w_iter + 1):
        for hi in range(h_iter + 1):
            np_img = coloring(np_img, wi, hi, check_w,
                              check_h, color_n, color1, color2)
            color_n += 1
    return np_img


def generate_gradation_texture(size_wh=(300, 300)):
    w, h = list(map(int, size_wh))
    np_img = np.zeros([w, h, 3]).astype(np.uint8)
    color1 = [np.random.randint(0, 255), np.random.randint(
        0, 255), np.random.randint(0, 255)]
    color2 = [np.random.randint(0, 255), np.random.randint(
        0, 255), np.random.randint(0, 255)]
    color3 = [np.random.randint(0, 255), np.random.randint(
        0, 255), np.random.randint(0, 255)]
    w_colors = np.array(
        [np.interp(np.arange(w), [0, w],
                   [color1[i], color2[i]]) for i in range(3)])
    h_colors = np.array(
        [np.interp(np.arange(h), [0, h],
                   [color1[i], color3[i]]) for i in range(3)])

    for wi in range(w):
        for hi in range(h):
            c = (w_colors[:, wi] + h_colors[:, hi]) / 2
            c = c.astype(np.uint8)
            np_img[wi, hi] = c
    return np_img


def generate_synthetic_texture(size_wh=(300, 300), style=None):
    valid_style = ['single', 'gradation', 'check']
    if style is None:
        style = random.choice(
            valid_style)

    if style == 'single':
        return generate_single_color_texture(size_wh)
    elif style == 'check':
        return generate_check_texture(size_wh)
    elif style == 'gradation':
        return generate_gradation_texture(size_wh)
    else:
        raise RuntimeError(f'Invalid style, {style}.'
                           + f'Valid styles are {valid_style}.'
                           )
